/*
 * UAE - The Un*x Amiga Emulator
 *
 * Hardfile emulation
 *
 * Copyright 1995 Bernd Schmidt
 */

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "threaddep/penguin.h"
#include "options.h"
#include "memory.h"
#include "custom.h"
#include "newcpu.h"
#include "disk.h"
#include "autoconf.h"
#include "filesys.h"
#include "execlib.h"
#include "native2amiga.h"

#include <dlfcn.h>
#include <devctl.h>
#include <sys/dcmd_cam.h>
#include <sys/cpt.h>

#define THREADED
#define THREAD_GO_DOWN 1

static int opencount = 0;

static uae_u32
hardfile_open(void)
{
	uaecptr tmp1 = m68k_areg(regs, 1);	/* IOReq */

	printf("hardfile open: ioreq = 0x%08lx, unit = %d, flags = %d\n",m68k_areg(regs,1),m68k_dreg(regs,0),m68k_dreg(regs,1));

	/* Check unit number */
	if (get_hardfile_data(m68k_dreg(regs, 0)))
	{
		opencount++;
		put_word(m68k_areg(regs, 6) + 32, get_word(m68k_areg(regs, 6) + 32) + 1);
		put_long(tmp1 + 24, m68k_dreg(regs, 0));	/* io_Unit */
		put_byte(tmp1 + 31, 0);	/* io_Error */
		put_byte(tmp1 + 8, 7);	/* ln_type = NT_REPLYMSG */
		return 0;
	}

	put_long(tmp1 + 20, (uae_u32) - 1);
	put_byte(tmp1 + 31, (uae_u8) - 1);
	return (uae_u32) - 1;
}

static uae_u32
hardfile_close(void)
{
	opencount--;
	put_word(m68k_areg(regs, 6) + 32, get_word(m68k_areg(regs, 6) + 32) - 1);

	return 0;
}

static uae_u32
hardfile_expunge(void)
{
	return 0;		/* Simply ignore this one... */
}

static void
check_change(struct hardfiledata * hfd)
{
	if (hfd->fd >= 0)
	{
		if (lseek(hfd->fd, 0, SEEK_CUR) < 0)
		{
			write_log("check_change: media removed\n");
			close(hfd->fd);
			hfd->fd = -1;
			hfd->size = 0;
			hfd->nrcyls = 0;
		}
	}

	if (hfd->fd == -1)
	{
		hfd->fd = open(hfd->unitinfo->rootdir, O_RDWR);
		if (hfd->fd >= 0)
		{
			hfd->unitinfo->readonly = 0;
		}
		else
		{
			hfd->fd = open(hfd->unitinfo->rootdir, O_RDONLY);
			if (hfd->fd >= 0)
				hfd->unitinfo->readonly = 1;
		}
		if (hfd->fd >= 0)
		{
			hfd->size = 0;
			write_log("check_change: fd opened\n");
		}
	}

	if (hfd->fd >= 0 && hfd->size == 0)
	{
		struct stat statbuf;

		if (stat(hfd->unitinfo->rootdir, &statbuf) != 0)
		{
			write_log("check_change: stat() failed: %s", strerror(errno));
			return;
		}
		if (statbuf.st_nblocks == 0)
		{
			write_log("check_change: st_nblocks == 0\n");
			return;
		}
		hfd->blocksize = statbuf.st_blocksize;
		hfd->size = statbuf.st_nblocks * hfd->blocksize;
		hfd->nrcyls = statbuf.st_nblocks / (hfd->secspertrack * hfd->surfaces);
		write_log("check change: media inserted, hfd->size = %d, hfd->nrcyls = %d\n", hfd->size, hfd->nrcyls);
	}
}


static uae_u32
handle_request(uae_u32 tmp1)
{
	uae_u32 tmp2, dataptr, offset, len;
	uae_u32 retval = m68k_dreg(regs, 0);
	int unit;
	struct hardfiledata *hfd;

	unit = get_long(tmp1 + 24);
	hfd = get_hardfile_data(unit);

	put_byte(tmp1 + 8, NT_MESSAGE);
	put_byte(tmp1 + 31, 0);	/* no error yet */
	tmp2 = get_word(tmp1 + 28);	/* io_Command */

	switch (tmp2)
	{
		case CMD_READ:
			dataptr = get_long(tmp1 + 40);
			if (dataptr & 1)
				goto bad_command;
			offset = get_long(tmp1 + 44);
			if (offset & 511)
				goto bad_command;
			tmp2 = get_long(tmp1 + 36);	/* io_Length */
			if (tmp2 & 511)
				goto bad_command;

			if (hfd->fd == -1 || hfd->size == 0)
				check_change(hfd);

			if (tmp2 + offset > (uae_u32) hfd->size)
			{
				printf("CMD_READ: size mismatch.\n");
				//goto bad_command;
			}

			if (hfd->fd == -1 || hfd->size == 0)
				goto bad_readcommand;

			put_long(tmp1 + 32, tmp2);	/* set io_Actual */
			if (lseek(hfd->fd, offset, SEEK_SET) == -1)
			{
				printf("CMD_READ: lseek() failed.\n");
				goto bad_readcommand;
			}
			while (tmp2)
			{
#ifdef FASTERIO
				if(read(hfd->fd, get_real_address(dataptr), tmp2) != tmp2)
					goto bad_readcommand;
				tmp2 = 0;
#else
				int i;
				char buffer[512];

				if (read(hfd->fd, buffer, 512) != 512) // could be interrupted legally ?
					goto bad_readcommand;
				for (i = 0; i < 512; i++, dataptr++)
					put_byte(dataptr, buffer[i]);
				tmp2 -= 512;
#endif
			}
			break;

bad_readcommand:
			write_log("bad_readcommand\n");
			check_change(hfd);
			put_byte(tmp1 + 31, (uae_u8) 29);	/* io_Error */
			retval = 0;
			break;

		case CMD_WRITE:
		case 11:	/* TD_FORMAT */
			dataptr = get_long(tmp1 + 40);
			if (dataptr & 1)
				goto bad_command;
			offset = get_long(tmp1 + 44);
			if (offset & 511)
				goto bad_command;
			tmp2 = get_long(tmp1 + 36);	/* io_Length */
			if (tmp2 & 511)
				goto bad_command;

			if (hfd->fd == -1 || hfd->size == 0)
				check_change(hfd);

			if (tmp2 + offset > (uae_u32) hfd->size)
				goto bad_command;

			if (hfd->fd == -1 || hfd->size == 0)
				goto bad_command;

			put_long(tmp1 + 32, tmp2);	/* set io_Actual */
			if (lseek(hfd->fd, offset, SEEK_SET) == -1)
				goto bad_command;
			while (tmp2)
			{
#ifdef FASTERIO
				if(write(hfd->fd, get_real_address(dataptr), tmp2) != tmp2)
					goto bad_command;
				tmp2 = 0;
#else
				char buffer[512];
				int i;

				for (i = 0; i < 512; i++, dataptr++)
					buffer[i] = get_byte(dataptr);
				if (write(hfd->fd, buffer, 512) != 512) // could be interrupted legally
					goto bad_command;
				tmp2 -= 512;
#endif
			}
			break;

	bad_command:
			break;

		case 14:	/* TD_CHANGESTATE */
			check_change(hfd);
			/* return 1 if no media is in drive */
			put_long(tmp1 + 32, (hfd->fd == -1 || hfd->size == 0) ? 1 : 0);
			break;

		case 18:	/* GetDriveType */
			write_log("GetDriveType: Shouldn't happen\n");
			put_long(tmp1 + 32, 1);	/* not exactly a 3.5" drive, but... */
			break;

		case 19:	/* GetNumTracks */
			put_long(tmp1 + 32, hfd ? hfd->nrcyls * hfd->surfaces : 0);
			break;

			/* Some commands that just do nothing and return zero */
		case CMD_UPDATE:
		case CMD_CLEAR:
		case 9: 	/* Motor */
		case 10:	/* Seek */
		case 12:	/* Remove */
		case 13:	/* ChangeNum */
		case 20:	/* AddChangeInt */
		case 21:	/* RemChangeInt */
			write_log("cmd = %d\n", tmp2);
			put_long(tmp1 + 32, 0);	/* io_Actual */
			retval = 0;
			break;

		case 15:	/* TD_PROTSTATUS */
			put_long(tmp1 + 32, hfd->unitinfo->readonly);	/* io_Actual */
			break;

		default:
			/* Command not understood. */
			write_log("unknown cmd = %d\n", tmp2);
			put_byte(tmp1 + 31, (uae_u8) - 3);	/* io_Error */
			retval = 0;
			break;
	}
	if ((get_byte(tmp1 + 30) & 1) == 0)
	{
		/* Not IOF_QUICK -- need to ReplyMsg */
#ifdef THREADED
		uae_ReplyMsg(tmp1);
#else
		m68k_areg(regs, 1) = tmp1;
		CallLib(get_long(4), -378);
#endif
	}
	return retval;
}

static int pipe_count = 0;

static uae_u32
hardfile_beginio(void)
{
	uae_u32 tmp1;
	UnitInfo *ui;

	tmp1 = m68k_areg(regs, 1);
	put_byte(tmp1 + 8, NT_MESSAGE);
#ifdef THREADED
	//no quick I / O, clear IOF_QUICK
	put_byte(tmp1 + 8, NT_MESSAGE);
	put_byte(tmp1 + 30, get_byte(tmp1 + 30) & ~1);
	ui = get_hardfile_unit(get_long(tmp1 + 24));
	if (!ui)
	{
		printf("hardfile_beginio: big trouble. unknown unit %d\n", get_long(tmp1 + 24));
		return 0;
	}
	//should never happen
	write_comm_pipe_pvoid(ui->unit_pipe, (void *) tmp1, 1);
	return 0;
#else
	return handle_request(tmp1);
#endif
}

static uae_u32
hardfile_abortio(void)
{
	return (uae_u32) - 3;
}

#ifdef THREADED
static void *
hardfile_penguin(void *unit_v)
{
	UnitInfo *ui = (UnitInfo *) unit_v;

	for (;;)
	{
		uae_u32 tmpl;

		tmpl = (uae_u32) read_comm_pipe_pvoid_blocking(ui->unit_pipe);

		if (ui->reset_state == THREAD_GO_DOWN)
		{
			sem_post(&ui->reset_sync_sem);
			return 0;
		}
		handle_request(tmpl);
	}
}

#endif

void
hardfile_initthread(UnitInfo * ui)
{
#ifdef THREADED
	int spolicy;
	struct sched_param sparam;

	check_change(&ui->hf);
	ui->unit_pipe = (smp_comm_pipe *) xmalloc(sizeof(smp_comm_pipe));
	init_comm_pipe(ui->unit_pipe, 50, 1);
	start_penguin(hardfile_penguin, (void *) (ui), &ui->tid);
	if (pthread_getschedparam(ui->tid, &spolicy, &sparam) == EOK)
	{
		sparam.sched_priority++;
		if (pthread_setschedparam(ui->tid, spolicy, &sparam) == EOK)
		{
			write_log("*** hardfile thread successfully configured.\n");
		}
	}
#endif
}

void
hardfile_prepare_reset(UnitInfo * ui)
{
#ifdef THREADED
	if (ui->unit_pipe)
	{
		sem_init(&ui->reset_sync_sem, 0, 0);
		ui->reset_state = THREAD_GO_DOWN;
		/* send death message */
		write_comm_pipe_int(ui->unit_pipe, 0, 1);
		sem_wait(&ui->reset_sync_sem);
	}
#endif
}

void
hardfile_install(void)
{
	uae_u32 functable, datatable;
	uae_u32 initcode, openfunc, closefunc, expungefunc;
	uae_u32 beginiofunc, abortiofunc;

	ROM_hardfile_resname = ds("axlb.device");
	ROM_hardfile_resid = ds("AXL block device 0.3");

	/* initcode */
#if 0
	initcode = here();
	calltrap(deftrap(hardfile_init));
	dw(RTS);
#else
	initcode = filesys_initcode;
#endif
	/* Open */
	openfunc = here();
	calltrap(deftrap(hardfile_open));
	dw(RTS);

	/* Close */
	closefunc = here();
	calltrap(deftrap(hardfile_close));
	dw(RTS);

	/* Expunge */
	expungefunc = here();
	calltrap(deftrap(hardfile_expunge));
	dw(RTS);

	/* BeginIO */
	beginiofunc = here();
	calltrap(deftrap(hardfile_beginio));
#if 0
	dw(0x48E7);
	dw(0x8002);		/* movem.l d0/a6,-(a7) */
	dw(0x0829);
	dw(0);
	dw(30);			/* btst #0,30(a1) */
	dw(0x6608);		/* bne.b +8 */
	dw(0x2C78);
	dw(0x0004);		/* move.l 4,a6 */
	dw(0x4EAE);
	dw(-378);		/* jsr ReplyMsg(a6) */
	dw(0x4CDF);
	dw(0x4001);		/* movem.l (a7)+,d0/a6 */
#endif
	dw(RTS);

	/* AbortIO */
	abortiofunc = here();
	calltrap(deftrap(hardfile_abortio));
	dw(RTS);

	/* FuncTable */
	functable = here();
	dl(openfunc);		/* Open */
	dl(closefunc);		/* Close */
	dl(expungefunc);	/* Expunge */
	dl(EXPANSION_nullfunc);	/* Null */
	dl(beginiofunc);	/* BeginIO */
	dl(abortiofunc);	/* AbortIO */
	dl(0xFFFFFFFF);	/* end of table */

	/* DataTable */
	datatable = here();
	dw(0xE000);		/* INITBYTE */
	dw(0x0008);		/* LN_TYPE */
	dw(0x0300);		/* NT_DEVICE */
	dw(0xC000);		/* INITLONG */
	dw(0x000A);		/* LN_NAME */
	dl(ROM_hardfile_resname);
	dw(0xE000);		/* INITBYTE */
	dw(0x000E);		/* LIB_FLAGS */
	dw(0x0600);		/* LIBF_SUMUSED | LIBF_CHANGED */
	dw(0xD000);		/* INITWORD */
	dw(0x0014);		/* LIB_VERSION */
	dw(0x0004);		/* 0.4 */
	dw(0xD000);
	dw(0x0016);		/* LIB_REVISION */
	dw(0x0000);
	dw(0xC000);
	dw(0x0018);		/* LIB_IDSTRING */
	dl(ROM_hardfile_resid);
	dw(0x0000);		/* end of table */

	ROM_hardfile_init = here();
	dl(0x00000100);		/* ??? */
	dl(functable);
	dl(datatable);
	dl(initcode);
}
