#ifndef EVENTS_H
#define EVENTS_H
 /*
  * UAE - The Un*x Amiga Emulator
  *
  * Events
  * These are best for low-frequency events. Having too many of them,
  * or using them for events that occur too frequently, can cause massive
  * slowdown.
  *
  * Copyright 1995-1998 Bernd Schmidt
  */

#include "machdep/rpt.h"


#if 0
extern volatile frame_time_t vsynctime, vsyncmintime;
extern void reset_frame_rate_hack (void);
extern int rpt_available;

/* Let's see whether hiding this away somewhere where the compiler can't
   see it will cure it of its silly urge to mis-optimize the comparison */
extern long int diff32(frame_time_t x, frame_time_t y);


extern unsigned long int event_cycles, nextevent, is_lastline;
extern unsigned long int sample_evtime;
typedef void (*evfunc)(void);

struct ev
{
    int active;
    unsigned long int evtime, oldcycles;
    evfunc handler;
};

enum {
    ev_hsync, ev_copper, ev_cia, ev_blitter, ev_disk,
    ev_max
};

extern struct ev eventtab[ev_max];

STATIC_INLINE void events_schedule (void)
{
    int i;

    unsigned long int mintime = ~0L;
    for(i = 0; i < ev_max; i++) {
	if (eventtab[i].active) {
	    unsigned long int eventtime = eventtab[i].evtime - event_cycles;
	    if (eventtime < mintime)
		mintime = eventtime;
	}
    }
    nextevent = event_cycles + mintime;
}

extern signed long pissoff;


STATIC_INLINE cycles_do_special(void)
{
    if (pissoff>=0)
	pissoff=-1;
}

STATIC_INLINE void do_extra_cycles (unsigned long cycles_to_add)
{
    pissoff-=cycles_to_add;
}

STATIC_INLINE unsigned long int get_cycles(void)
{
    return event_cycles;
}

STATIC_INLINE void init_cycles(void)
{
    event_cycles=0;
    pissoff=0;
}

STATIC_INLINE void do_cycles_slow (unsigned long cycles_to_add)
{
    if ((pissoff-=cycles_to_add)>=0)
	return;

    cycles_to_add=-pissoff;
    pissoff=0;
    
    if (is_lastline && eventtab[ev_hsync].evtime-event_cycles <= cycles_to_add) {
	frame_time_t now=read_processor_time();
	if (diff32(now, vsyncmintime)<0)
	{
	    pissoff=3000;
	    return;
	}
    }
    if ((nextevent - event_cycles) <= cycles_to_add) {
	for (; cycles_to_add != 0; cycles_to_add--) {
	    if (++event_cycles == nextevent) {
		int i;

		for (i = 0; i < ev_max; i++) {
		    if (eventtab[i].active && eventtab[i].evtime == event_cycles) {
			(*eventtab[i].handler)();
		    }
		}
		events_schedule();
	    }
	}
    }
    else {
	event_cycles += cycles_to_add;
    }
}

#define do_cycles do_cycles_slow
#define countdown pissoff

#else

extern volatile frame_time_t vsynctime, vsyncmintime;
extern void reset_frame_rate_hack (void);
extern int rpt_available;

/* Let's see whether hiding this away somewhere where the compiler can't
   see it will cure it of its silly urge to mis-optimize the comparison */
extern long int diff32(frame_time_t x, frame_time_t y);


extern unsigned long int event_cycles, nextevent, is_lastline;
extern long cycles_to_next_event;
extern long max_cycles_to_next_event;
extern long cycles_to_hsync_event;
extern long pissoff;

extern unsigned long int sample_evtime;
typedef void (*evfunc)(void);

struct ev
{
    int active;
    unsigned long int evtime, oldcycles;
    evfunc handler;
};

enum {
    ev_hsync, ev_copper, ev_cia, ev_blitter, ev_disk,
    ev_max
};

extern struct ev eventtab[ev_max];

STATIC_INLINE void cycles_do_special(void)
{
    if (cycles_to_next_event>=0) {
	pissoff+=cycles_to_next_event;
	cycles_to_next_event=0;
    }
}


STATIC_INLINE void do_extra_cycles (unsigned long cycles_to_add)
{
    pissoff-=cycles_to_add;
}

STATIC_INLINE unsigned long int get_cycles(void)
{
    if (cycles_to_next_event<=max_cycles_to_next_event) {
	max_cycles_to_next_event=cycles_to_next_event;
	return nextevent-cycles_to_next_event;
    }
    else {
	return nextevent-max_cycles_to_next_event;
    }
}

STATIC_INLINE void events_schedule (void)
{
    int i;

    unsigned long int curcycles=get_cycles();
    unsigned long int mintime = ~0L;

    cycles_to_hsync_event=eventtab[ev_hsync].evtime - curcycles;
    for(i = 0; i < ev_max; i++) {
	if (eventtab[i].active) {
	    unsigned long int eventtime = eventtab[i].evtime - curcycles;
	    if (eventtime < mintime) 
		mintime = eventtime;
	}
    }
    nextevent = curcycles + mintime;
    max_cycles_to_next_event=mintime;
    cycles_to_next_event=mintime;
    cycles_to_hsync_event=mintime-cycles_to_hsync_event;
}

STATIC_INLINE void init_cycles(void)
{
    int i;

    /* This only happens just before the whole table gets blattered, anyway,
       so we can do nasty things here */
    nextevent=cycles_to_next_event=max_cycles_to_next_event=0;
    pissoff=0;

    /* This is copied here from custom.c. I think it really belongs in here;
       When the cycles get reset to 0, all the events should be disabled! */
    for(i = 0; i < ev_max; i++) {
	eventtab[i].active = 0;
	eventtab[i].oldcycles = 0;
    }
}

STATIC_INLINE void do_cycles_slow (long cycles_to_add)
{
    cycles_to_next_event-=cycles_to_add;
    if (cycles_to_next_event>0)
	return;

    cycles_to_next_event+=pissoff;
    pissoff=0;

    if (is_lastline && 
	cycles_to_next_event <= cycles_to_hsync_event) {
	frame_time_t now=read_processor_time();
	if (diff32(now, vsyncmintime)<0)
	{
	    cycles_to_next_event+=3000;
	    return;
	}
    }
    cycles_to_add=0;
    while (cycles_to_next_event <= cycles_to_add) {
	int i;

	cycles_to_add-=cycles_to_next_event;
	cycles_to_next_event=0;
	for (i = 0; i < ev_max; i++) {
	    if (eventtab[i].active && eventtab[i].evtime == nextevent) {
		(*eventtab[i].handler) ();
	    }
	}
	events_schedule();
    }
    cycles_to_next_event-=cycles_to_add;
}

#define do_cycles do_cycles_slow
#define countdown cycles_to_next_event

#endif

#endif
