 /*
  * UAE - The Un*x Amiga Emulator
  *
  * Call (some) Amiga Exec functions outside the main UAE thread
  * and without stack magic.
  *
  * Copyright 1999 Patrick Ohly
  *
  * Uses the EXTER interrupt that is setup in filesys.c
  * and some of it needs thread support.
  */

/*
 * The following functions do exactly the same thing as their
 * Amiga counterpart, but can be called in situation where calling
 * the exec.library functions is impossible.
 */
#ifdef SUPPORT_THREADS
void uae_ReplyMsg(uaecptr msg);
void uae_PutMsg(uaecptr port, uaecptr msg);
uaecptr uae_GetMsg(uaecptr port);
void uae_Signal(uaecptr task, uae_u32 mask);
#endif
void uae_NewList(uaecptr list);

/*
 * The following functions are shortcuts for calling
 * the exec.library function with CallLib(), so they
 * are only available in a trap function. This trap
 * function has to be setup with deftrap2() and
 * TRAPFLAG_EXTRA_STACK and stack magic is required.
 */
uaecptr uae_AllocMem (uae_u32 size, uae_u32 flags);
void uae_FreeMem(uaecptr mem, uae_u32 size);
uaecptr uae_AllocVec(uae_u32 size, uae_u32 flags);
void uae_FreeVec(uaecptr mem);

/*
 * to be called when setting up the hardware
 */
void native2amiga_install (void);

/*
 * to be called when the Amiga boots, i.e. by filesys_diagentry()
 */
void native2amiga_startup (void);

/*

	This is a second API that allows to call Amiga Library Functions from
	a thread.

   pipe: the communication pipe
   libbase: struct Library *
   funcoffset: offset of library function (negative value)
   retval: return value from function or undefined
   dregs: D0 - D7 on call, undefined after return
   aregs: A0 - A7, but A6 and A7 not used, undefined after return
   msg: the Amiga message (private structure)
*/

typedef struct _AOSLibCallPkt
{
    smp_comm_pipe pipe;
    uaecptr libbase;
    uae_u32 funcoffset;
    uae_u32 retval;
    uae_u32 dregs[8];
    uae_u32 aregs[8];
    uaecptr msg;
    int msgused;
    int msgsize;
} AOSLibCallPkt;

/* you must allocate a call packet with this function. This function may
   wait on a semaphore until a package is available.
*/
AOSLibCallPkt *get_aoslibcallpkt(void);

/* allocate some native data from the packet. returns NULL if not enough
   memory is free or an address in the amiga address space that can be
   transferred to the Amiga using one of the registers or within a
   structure that were copied before for the same packet.
*/
uaecptr alloc_aoslibcalldata(AOSLibCallPkt *pkt, int len);

/* copy some native data into the packet. returns NULL if not enough
   memory is free or an address in the amiga address space that can be
   transferred to the Amiga using one of the registers or within a
   structure that were copied before for the same packet.
*/
uaecptr copy_aoslibcalldata(AOSLibCallPkt *pkt, void *data, int len);

/* free a previously allocated packet.
*/
void free_aoslibcallpkt(AOSLibCallPkt *pkt);

/* call the function from a thread only. Returns 0 if the called failed - the
   function result is in dregs[0] (M68K register D0).
*/
int call_aoslib_from_thread(AOSLibCallPkt *pkt);

/* call a AmigaOS function. This call automatically differs between calls
   from a thread and calls from the emulator itself - which assumes that
   the native code was called using a trap.
*/
int call_aoslib(AOSLibCallPkt *pkt);

/**** internal stuff ****/
#ifdef SUPPORT_THREADS
/* This pipe is filled by Signal() with pairs of
 * (uae_u32)0/(uaecptr)task/(uae_u32)signal_set,
 * by PutMsg() with (uae_u32)1/(uaecptr)port/(uaecptr)msg and by
 * ReplyMsg() with (uae_u32)2/(uaecptr)msg.
 * It's emptied via exter_int_helper by the EXTER interrupt. */
extern smp_comm_pipe native2amiga_pending;
#endif
