typedef unsigned int ULONG;
#define VOID void

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"
#include "custom.h"
#include "newcpu.h"
#include "autoconf.h"
#include "threaddep/penguin.h"
#include "native2amiga.h"

//#include "libraries/native.h"
#include "nativelib.h"
#include <pthread.h>
#include <signal.h>
#include <dlfcn.h>
#include <stdio.h>
#include <stdlib.h>

#ifndef NDEBUG
#define DEBUG(x)
#else
#define DEBUG(x)
#endif

struct NativePenguin
{
	struct NativePenguin *next;
	struct NativePenguin *prev;
	penguin_id id;
	smp_comm_pipe commpipe;
	int canceled;
};


static pthread_mutex_t firstProtect = PTHREAD_MUTEX_INITIALIZER;
static struct NativePenguin *firstNativePenguin;


static void nativelib_cleanup(void)
{
	struct NativePenguin *next, *walk = firstNativePenguin;
	DEBUG(printf("nativelib_cleanup: entering...\n");)
	while(walk)
	{
		next = walk->next;
		if(walk->canceled)
		{
			pthread_join(walk->id, NULL);
			if(walk->prev)
				walk->prev->next = walk->next;
			else
				firstNativePenguin = walk->next;
			if(walk->next)
				walk->next->prev = walk->prev;
			destroy_comm_pipe(&walk->commpipe);
			free(walk);
			DEBUG(printf("nativelib_cleanup: thread removed.\n");)
		}
		walk = next;
	}
	DEBUG(printf("nativelib_cleanup: done.\n");)
}


static void signal_ignorer(int s)
{
	//printf("*** Ignoring signal %d.\n", s);
	signal(s, signal_ignorer);
}


static void *nativelib_penguin(void *acontext)
{
	int s;
	void *a, *v;
	void *(*f)(void*);
	struct NativePenguin *comm = (struct NativePenguin*) get_long((int)acontext);
	uaecptr amiga_task = (uaecptr) get_long(((int)acontext) + 4);
	uae_u32 amiga_sigm = (uae_u32) get_long(((int)acontext) + 8);
	sigset_t sigblock;

	//printf("*** nativelib_penguin ready. amiga_task = %p, amiga_signalmask = 0x%08x\n", amiga_task, amiga_sigm);
	if(sigemptyset(&sigblock) == 0)
	{
		if(sigaddset(&sigblock, 32) != 0)
			printf("*** nativelib_penguin: sigaddset() failed.\n");
	}
	else printf("*** nativelib_penguin: sigemptyset() failed.\n");

	pthread_sigmask(SIG_BLOCK, &sigblock, NULL);
	uae_Signal(amiga_task, amiga_sigm);
	while(1)
	{
		pthread_testcancel();
		s = read_comm_pipe_int_blocking(&comm->commpipe);
		a = read_comm_pipe_pvoid_blocking(&comm->commpipe);
		//printf("*** nativelib_penguin: id = %d, received switch = %d, arg = %p.\n", pthread_self(), s, a);
		switch(s)
		{
			case NATIVE_OPENOBJECT:
				DEBUG(printf("nativelib_penguin: NATIVE_OPENOBJECT '%s'.\n", (const char*)get_real_address(get_long(((int)a) + 4)));)
				put_long(((int)a) + 8, (int)(v = dlopen((const char*)get_real_address(get_long(((int)a) + 4)), RTLD_NOW)));
				/*if(!v)
					printf("*** nativelib_penguin failed to open object: %s\n", dlerror());*/
				break;

			case NATIVE_CLOSEOBJECT:
				DEBUG(printf("nativelib_penguin: NATIVE_CLOSEOBJECT\n");)
				//dlclose((void*)get_long(((int)a) + 8));
				put_long(((int)a) + 8, NULL);
				break;

			case NATIVE_GETSYMBOLID:
				//DEBUG(printf("nativelib_penguin: NATIVE_GETSYMBOLID\n");)
				//printf("*** nativelib_penguin: symbol = '%s', 0x%08x, object = 0x%08x.\n", (const char*)get_real_address(get_long(((int)a) + 4)), get_long(((int)a) + 4), get_long(get_long((int)a) + 8));
				put_long(((int)a) + 8, (int)dlsym((void*)get_long(get_long((int)a) + 8), (const char*)get_real_address(get_long(((int)a) + 4))));
				break;

			case NATIVE_CALLFUNCTION:
				//DEBUG(printf("nativelib_penguin: NATIVE_CALLFUNCTION\n");)
				pthread_sigmask(SIG_UNBLOCK, &sigblock, NULL);
				f = (void*(*)(void*))get_long(((int)a) + 4);
				//printf("*** nativelib_penguin: function = 0x%08x.\n", f);
				put_long(((int)a) + 12, (int)(f((void*)get_long(((int)a) + 8))));
				pthread_sigmask(SIG_BLOCK, &sigblock, NULL);
				break;

			case NATIVE_SHUTDOWN:
				DEBUG(printf("nativelib_penguin: NATIVE_SHUTDOWN\n");)
				return 0;
		}
		//printf("*** nativelib_penguin: done. sending signal...\n");
		uae_Signal(amiga_task, amiga_sigm);
		//printf("*** nativelib_penguin: Ok.\n");
	}
	return 0;
}


static uae_u32 nativelib_handler(void)
{
	/* returns 0: Don't expect a signal. Maybe an error occured.
	** returns 1: request successfully posted. Expect signal on completion. */
	struct NativePenguin *comm;
	int s = m68k_dreg(regs, 0);
	int a = (int) m68k_areg(regs, 0);
	DEBUG(printf("nativelib_handler: entering...\n");)
	switch(s)
	{
		case NATIVE_CREATECONTEXT:
			DEBUG(printf("nativelib_handler: NATIVE_CREATECONTEXT (Amiga-Task = '%s')\n", (char*)get_real_address(get_long(get_long(a + 4) + 10)));)
			//nativelib_cleanup();
			comm = (struct NativePenguin*) malloc(sizeof(struct NativePenguin));
			if(!comm)
			{
				DEBUG(printf("nativelib_handler: NATIVE_CREATECONTEXT failed.\n");)
				return 0;
			}
			bzero(comm, sizeof(struct NativePenguin));
			init_comm_pipe(&comm->commpipe, 8, 2);
			put_long(a, (int)comm);	/* NativeContext->native_comm */
			if(start_penguin(nativelib_penguin, (void*)a, &comm->id) != EOK)
			{
				put_long(a, NULL);	/* NativeContext->native_comm */
				free(comm);
				DEBUG(printf("nativelib_handler: NATIVE_CREATECONTEXT failed (no more penguins).\n");)
				return 0;
			}
			else
			{
				int spolicy;
				struct sched_param sparam;
			    if(pthread_getschedparam(comm->id, &spolicy, &sparam) == EOK)
	    		{
			    	sparam.sched_priority ++;
	    			pthread_setschedparam(comm->id, spolicy, &sparam);
		    	}
		    }
			comm->next = firstNativePenguin;
			if(comm->next)
				comm->next->prev = comm;
			firstNativePenguin = comm;
			DEBUG(printf("nativelib_handler: NATIVE_CREATECONTEXT done.\n");)
			return 1;

		case NATIVE_DESTROYCONTEXT:
			DEBUG(printf("nativelib_handler: NATIVE_DESTROYCONTEXT (Amiga-Task = '%s')\n", (char*)get_real_address(get_long(get_long(a + 4) + 10)));)
			comm = (struct NativePenguin*) get_long(a);
			if(comm)
			{
				pthread_cancel(comm->id);
				write_comm_pipe_int(&comm->commpipe, NATIVE_SHUTDOWN, 0);
				write_comm_pipe_pvoid(&comm->commpipe, NULL, 0);
				comm->canceled = 1;
			}
			DEBUG(printf("nativelib_handler: NATIVE_DESTROYCONTEXT done.\n");)
			return 0;

		case NATIVE_OPENOBJECT:
		case NATIVE_CLOSEOBJECT:
			comm = (struct NativePenguin*) get_long(get_long(a));
			write_comm_pipe_int(&comm->commpipe, s, 0);
			write_comm_pipe_pvoid(&comm->commpipe, (void*)a, 0);
			return 1;

		case NATIVE_GETSYMBOLID:
		case NATIVE_CALLFUNCTION:
			comm = (struct NativePenguin*) get_long(get_long(get_long(a)));
			write_comm_pipe_int(&comm->commpipe, s, 0);
			write_comm_pipe_pvoid(&comm->commpipe, (void*)a, 0);
			return 1;

		case NATIVE_SIGNAL:
			comm = (struct NativePenguin*) get_long(get_long(get_long(a)));
			pthread_kill(comm->id, (int) get_long(a + 4));
			return 0;

		default:
			fprintf(stderr, "nativelib_handler(): unknown switch = %d, arg = %p\n", s, a);
	}
	return 0;
}


void nativelib_install(void)
{
	uaecptr inst;

	signal(32, signal_ignorer);
	signal(SIGPIPE, signal_ignorer);
	pthread_mutex_init(&firstProtect, NULL);
	firstNativePenguin = NULL;

	inst = here();
	org(0xF0FFC0);
	calltrap(deftrap(nativelib_handler));
	dw(RTS);
	org(inst);
}


void nativelib_exit(void)
{
	nativelib_reset();
	pthread_mutex_destroy(&firstProtect);
}


void nativelib_reset(void)
{
	struct NativePenguin *walk = firstNativePenguin;
	DEBUG(printf("nativelib_reset: entering...\n");)
	while(walk)
	{
		if(!walk->canceled)
		{
			pthread_cancel(walk->id);
			write_comm_pipe_int(&walk->commpipe, NATIVE_SHUTDOWN, 0);
			write_comm_pipe_pvoid(&walk->commpipe, NULL, 0);
			walk->canceled = 1;
		}
		walk = walk->next;
	}
	nativelib_cleanup();
	DEBUG(printf("nativelib_reset: done.\n");)
}

