/*
 * UAE - The Un*x Amiga Emulator
 *
 * Photon Direct (QNX) interface
 *
 * Copyright 2001 Jochen Becher
 *
 */

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"
#include "threaddep/penguin.h"
#include "uae.h"
#include "memory.h"
#include "xwin.h"
#include "custom.h"
#include "drawing.h"
#include "newcpu.h"
#include "keyboard.h"
#include "keybuf.h"
#include "gui.h"
#include "debug.h"
#include "picasso96.h"
#include "native2amiga.h"

#include "screenbar.h"
#define SCREENBAR

#define Pk_BASIC_LATIN
#define Pk_LATIN1
#include <Ph.h>
#include <Pt.h>
#include <photon/PhSystem.h>
#include <photon/PhCursor.h>
#define PX_IMAGE_MODULES
#define PX_GIF_SUPPORT
#define PX_JPG_SUPPORT
#define PX_PNG_SUPPORT
#include <photon/PxImage.h>

//extern int joy1_button1, joy1_button2, joy1_bot, joy1_right, joy1_top, joy1_left;

extern uaecptr AmigaServerTask;
#define SIG_AmigaToFront 8192 /* move special screen to back */
#define SIG_AmigaToBack 16384 /* move special screen to front */
#define SIG_SetMouse (1<<30) /* set mouse position */

#define udata_last_mouse_x 0
#define udata_last_mouse_y 4
#define udata_new_mouse_x 8
#define udata_new_mouse_y 12

#define tc_UserData 0x0058

void leave_graphics_mode(int noscreenbar);

static RETSIGTYPE sigbrkhandler (int foo)
{
#if 1
	leave_graphics_mode(1);
	exit(1);
#else
	leave_graphics_mode(1);
	activate_debugger();
	signal (SIGINT, sigbrkhandler);
#endif
}

static RETSIGTYPE sigtermhandler (int foo)
{
	leave_graphics_mode(1);
	exit(1);
}

void setup_brkhandler (void)
{
	signal (SIGFPE, SIG_IGN);
	signal (SIGINT, sigbrkhandler);
	signal (SIGTERM, sigtermhandler);
	signal (SIGQUIT, sigtermhandler);
//	signal (SIGTSTP, sigtermhandler);
	signal (SIGPWR, sigtermhandler);
}

/***************************************************************/

struct PicassoMode
{
	PgDisplaySettings_t mode;
	int width;
	int height;
	int depth;
	uae_u32 rgbformat;
	int scaledX;	/* the real width  of the mode */
	int scaledY;	/* the real height of the mode */
};

static PdDirectContext_t *DirectMode = NULL;
static PhDrawContext_t *Olddc = NULL;

static PgDisplaySettings_t OldVideoMode;
static int OldVideoModeValid = 0;

static int phmouse_x = 0, phmouse_y = 0;

static int amiga_in_front = 0;

#define AGAWIDTH 800
#define AGAHEIGHT 600
static PgDisplaySettings_t DirectAGAMode;
static char *AGAGfxmem = NULL;
static PdOffscreenContext_t *AGAVideoContext = NULL;
static char *AGAVideoRam = NULL;
static int AGAVideoPixBits = 0;
static int AGAVideoPixBytes = 0;
static int AGAVideoRowBytes = 0;

#define PICASSO_MAX_MODES MAX_PICASSO_MODES
static int DirectPicassoModeNum = 0;
static struct PicassoMode DirectPicassoModes[PICASSO_MAX_MODES];
static int LastDirectPicassoMode = -1;
static PdOffscreenContext_t *PicassoVideoContext = NULL;
static char *PicassoVideoRam = NULL;

static int screen_is_picasso = 0;
static char picasso_invalid_lines[10000];
static int picasso_has_invalid_lines = 0;
static int picasso_invalid_start, picasso_invalid_stop;
static int picasso_maxw = 0, picasso_maxh = 0;

/***************************************************************/

extern uae_u16 lof;
extern unsigned int bplcon0;

void flush_line (int y)
{
	//printf("flush_line %d\n",y);
	flush_block (y,y);
}

void flush_block (int ystart, int ystop)
{
#if 0
	static int repeat = 7;	// There seem to be 7 0-throughwalks of ystart until the Workbench Screen opens.

	if(repeat > 0)
	{
		if(repeat == 7)
		{
			PhImage_t *bootpic = PxLoadImage("boot.pic", NULL);
			if(bootpic)
			{
				PhPoint_t pos = { 0, 0 };
				PgDrawPhImagemx(&pos, bootpic, 0);
				PgFlush();
				PhReleaseImage(bootpic);
				repeat --;
			}
			else
			{
				repeat = 0;
			}
		}

		if(repeat > 0)
		{
			if(ystart == 0)
				repeat --;
			return;
		}
	}
#endif

	if (!amiga_in_front)
		return;
	if (ystart < 0 || ystop >= AGAHEIGHT)
		return;

	if (AGAVideoRam)
	{
		if (AGAVideoPixBytes == 4 || AGAVideoPixBytes == 2)
		{
			unsigned long long *from = (unsigned long long *) (gfxvidinfo.bufmem /*+ ystart * gfxvidinfo.rowbytes*/);
			unsigned long long *to = (unsigned long long *) (AGAVideoRam + ystart * AGAVideoRowBytes);
			int y;
			int n = gfxvidinfo.rowbytes / 8;
			for (y = ystart; y <= ystop; y++)
			{
				int i;
				for (i = 0; i < n; i++)
					*to++ = *from++;
			}
		}
		else if (AGAVideoPixBytes == 3)
		{
			char *from = (char *) (gfxvidinfo.bufmem /*+ ystart * gfxvidinfo.rowbytes*/);
			char *to = (char *) (AGAVideoRam + ystart * AGAVideoRowBytes);
			int y;
			int n = gfxvidinfo.rowbytes / 4;
			for (y = ystart; y <= ystop; y++)
			{
				int i;
				for (i = 0; i < n; i++)
				{
					*to++ = *from++;
					*to++ = *from++;
					*to++ = *from++;
					from++;
				}
			}
		}
		else
		{
		}
	}
	else
	{
		int type;
		PhPoint_t pos = { 0, ystart };
		PhDim_t size = { gfxvidinfo.rowbytes * 8, ystop - ystart + 1 };

		//printf("flush_block %d to %d\n",ystart,ystop);
		if (gfxvidinfo.pixbytes == 3)
			type = Pg_IMAGE_DIRECT_888;
		else if (gfxvidinfo.pixbytes == 4)
			type = Pg_IMAGE_DIRECT_8888;
		else if (gfxvidinfo.pixbytes == 2)
			type = Pg_IMAGE_DIRECT_565;
		else
		{
			write_log("Unsupported bytes per pixel\n");
			return;
		}
		PgDrawImage(gfxvidinfo.bufmem /* + ystart * gfxvidinfo.rowbytes */,type,&pos,&size,gfxvidinfo.rowbytes,0);
		PgFlush();
	}
}

void flush_screen (int ystart, int ystop)
{
}

static void flush_block_picasso(int ystart, int ystop)
{
	char *addr;

	if (!amiga_in_front)
		return;
	addr = gfxmemory
		+ (picasso96_state.Address - gfxmem_start)
		+ ystart * picasso96_state.BytesPerRow;
	if (PicassoVideoRam)
	{
		int sx = DirectPicassoModes[LastDirectPicassoMode].scaledX
		       / DirectPicassoModes[LastDirectPicassoMode].width;
		int sy = DirectPicassoModes[LastDirectPicassoMode].scaledY
		       / DirectPicassoModes[LastDirectPicassoMode].height;

		//printf("flush_block_picasso: sx = %d, sy = %d, bpp = %d, bpr = %d, addr = 0x%08x.\n", sx, sy, picasso_vidinfo.pixbytes, picasso_vidinfo.rowbytes, addr);
		if(sx < 2  &&  sy < 2)
		{
			int x;
			int llpr = picasso_vidinfo.rowbytes / 8;
			int fllpr = (picasso96_state.BytesPerRow - picasso_vidinfo.rowbytes) / 8;
			unsigned long long *from = (unsigned long long *) addr;
			unsigned long long *to = (unsigned long long *) (PicassoVideoRam + ystart * picasso_vidinfo.rowbytes);
			while(ystart <= ystop)
			{
				for(x=llpr; x>0; x--)
					*to++ = *from++;
				from += fllpr;
				ystart ++;
			}
		}
		else if(sy == 2)
		{
			if(picasso_vidinfo.pixbytes == 1)
			{
				int s, x;
				int bpr = picasso_vidinfo.rowbytes;
				int fbpr = picasso96_state.BytesPerRow - picasso_vidinfo.rowbytes;
				unsigned char v;
				unsigned char *from = (unsigned char*) addr;
				unsigned char *to1  = (unsigned char*) (PicassoVideoRam + 2 * ystart * bpr * sx);
				unsigned char *to2  = to1 + bpr * sx;
				while(ystart <= ystop)
				{
					for(x=bpr; x>0; x--)
					{
						v = *from++;
						for(s=sx; s>0; s--)
						{
							*to1++ = v;
							*to2++ = v;
						}
					}
					from += fbpr;
					to1 += bpr * sx;
					to2 += bpr * sx;
					ystart ++;
				}
			}
			else if(picasso_vidinfo.pixbytes == 2)
			{
				int s, x;
				int spr = picasso_vidinfo.rowbytes / 2;
				int fspr = (picasso96_state.BytesPerRow - picasso_vidinfo.rowbytes) / 2;
				unsigned short v;
				unsigned short *from = (unsigned short*) addr;
				unsigned short *to1  = (unsigned short*) (PicassoVideoRam + 2 * ystart * spr * 2 * sx);
				unsigned short *to2  = to1 + spr * sx;
				while(ystart <= ystop)
				{
					for(x = spr; x > 0; x--)
					{
						v = *from++;
						for(s=sx; s>0; s--)
						{
							*to1++ = v;
							*to2++ = v;
						}
					}
					from += fspr;
					to1 += spr * sx;
					to2 += spr * sx;
					ystart ++;
				}
			}
			else if(picasso_vidinfo.pixbytes == 4)
			{
				int s, x;
				int lpr = picasso_vidinfo.rowbytes / 4;
				int flpr = (picasso96_state.BytesPerRow - picasso_vidinfo.rowbytes) / 4;
				unsigned long v;
				unsigned long *from = (unsigned long*) addr;
				unsigned long *to1  = (unsigned long*) (PicassoVideoRam + 2 * ystart * lpr * 4 * sx);
				unsigned long *to2  = to1 + lpr * sx;
				while(ystart <= ystop)
				{
					for(x = lpr; x > 0; x--)
					{
						v = *from++;
						for(s=sx; s>0; s--)
						{
							*to1++ = v;
							*to2++ = v;
						}
					}
					from += flpr;
					to1 += lpr * sx;
					to2 += lpr * sx;
					ystart ++;
				}
			}
		}
	}
	else
	{
		int type;
		PhPoint_t pos = { 0, ystart };
		PhDim_t size = { picasso_vidinfo.rowbytes * 8, ystop - ystart + 1 };

		if (picasso96_state.BytesPerPixel == 4 && picasso96_state.BytesPerRow * 8 >= 1024)
			write_log("flush_block_picasso %d to %d\n",ystart,ystop);

		if (picasso96_state.BytesPerPixel == 3)
			type = Pg_IMAGE_DIRECT_888;
		else if (picasso96_state.BytesPerPixel == 4)
			type = Pg_IMAGE_DIRECT_8888;
		else if (picasso96_state.BytesPerPixel == 2)
			type = Pg_IMAGE_DIRECT_565;
		else if (picasso96_state.BytesPerPixel == 1)
			type = Pg_IMAGE_PALETTE_BYTE;
		else
		{
			write_log("Unsupported bytes per pixel\n");
			return;
		}
		PgDrawImage(addr,type,&pos,&size,picasso96_state.BytesPerRow,0);
		PgFlush();
	}
}

/***************************************************************/

#define OPAQUEMASK Ph_EV_RAW

static PhRid_t input_region = (PhRid_t) -1;
static size_t eventsize = 0;
static PhEvent_t *event = NULL;
static int keystate[256]; // while amiga screen is in front
static int qnxkeystate[256]; // while qnx workspace is in front

static int open_region(void)
{
	int i;

	if (input_region == -1)
	{
		PhRegion_t info;
		PhRect_t rect = { { -32768, -32768 }, { 32767, 32767 } };

		bzero(&info,sizeof(info));

		info.events_sense = Ph_EV_RAW | Ph_EV_SYSTEM;
		info.events_opaque = 0; // OPAQUEMASK;
		info.parent = Ph_DEV_RID;
		info.cursor_type = Ph_CURSOR_CROSSHAIR;
		input_region = PhRegionOpen(Ph_REGION_EV_SENSE | Ph_REGION_EV_OPAQUE
				| Ph_REGION_PARENT /* | Ph_REGION_CURSOR*/ | Ph_REGION_RECT,
			&info, &rect, NULL);
		if (input_region == -1)
		{
			fprintf(stderr, "Unable to open input region\n");
			return 0;
		}
	}
	if (event == NULL)
	{
		eventsize = sizeof(PhEvent_t) + 1000;
		event = (PhEvent_t *) malloc(eventsize);
		if (!event)
		{
			fprintf(stderr,"out of mem for photon event\n");
			return 0;
		}
		for (i = 0; i < 256; i++)
		{
			keystate[i] = 0;
			qnxkeystate[i] = 0;
		}
	}
	return 1;
}

static void close_region(void)
{
	if (input_region != -1)
	{
		PhRegionClose(input_region);
		input_region = -1;
	}
	if (event)
	{
		free(event);
		event = NULL;
		eventsize = 0;
	}
}

static void transparent_region(void)
{
	if (input_region != -1)
	{
		PhRegion_t info;

		bzero(&info,sizeof(info));
		info.rid = input_region;
		info.events_opaque = 0;
		PhRegionChange(Ph_REGION_EV_OPAQUE,0,&info,NULL,NULL);
	}
}

static void opaque_region(void)
{
	if (input_region != -1)
	{
		PhRegion_t info;

		bzero(&info,sizeof(info));
		info.rid = input_region;
		info.events_opaque = OPAQUEMASK;
		PhRegionChange(Ph_REGION_EV_OPAQUE,0,&info,NULL,NULL);
	}
}

/****************************************************************/

struct RawKeyLevel0
{
	int scancode;
	int rawkeycode; // == <0 ? -> -index in level1 table
};

struct RawKeyLevel1
{
	int next; // 0 ? end of queue : index to next entry
	int rawkeycode;
	int numscan; // don't count first scan code from level 0
	int scancodes[1];
};

static struct RawKeyLevel0 rawkeytable[256];
static int rawkeyextents[512];
static int rawkeyextent;
static int keybuffer[10];
static int keybufferlen;

static int scan_key(FILE *keyfile, int rawkeycode, int relkey)
{
	int numcodes, scancode, i;

	if (fscanf(keyfile,"%i",&numcodes) != 1)
		return 0;
	if (numcodes == 1)
	{
		if (fscanf(keyfile,"%i",&scancode) != 1)
			return 0;
		rawkeytable[scancode].scancode = scancode;
		rawkeytable[scancode].rawkeycode = rawkeycode | (relkey ? 0x80 : 0x00);
	}
	else if (numcodes > 1)
	{
		if (fscanf(keyfile,"%i",&scancode) != 1)
			return 0;
		if (rawkeytable[scancode].rawkeycode >= 0
			&& rawkeytable[scancode].rawkeycode <= 255)
		{
			fprintf(stderr, "scan code 0x%02x is already used\nas single code for rawkey 0x%02x.\n",scancode,rawkeytable[scancode].rawkeycode);
			return 0;
		}
		rawkeytable[scancode].scancode = scancode;
		if (rawkeytable[scancode].rawkeycode > 255)
			rawkeytable[scancode].rawkeycode = -rawkeyextent;
		else
		{
			i = -rawkeytable[scancode].rawkeycode;
			while (rawkeyextents[i] != 0)
				i = rawkeyextents[i];
			rawkeyextents[i] = rawkeyextent;
		}
		rawkeyextents[rawkeyextent++] = 0;
		rawkeyextents[rawkeyextent++] = rawkeycode | (relkey ? 0x80 : 0x00);
		rawkeyextents[rawkeyextent++] = numcodes-1;
		for (i = 0; i < numcodes-1; i++)
		{
			if (fscanf(keyfile,"%i",&scancode) != 1)
				return 0;
			rawkeyextents[rawkeyextent++] = scancode;
		}
	}
	else
		return 0;
	return 1;
}

static void prepare_keytable(void)
{
	int i;

	for (i = 0; i < 255; i++)
	{
		rawkeytable[i].scancode = 0;
		rawkeytable[i].rawkeycode = 256; // positive and >255 -> not used
	}
	rawkeyextent = 1; // no key extents yet, 0 is reserved
}

static int load_keytable(void)
{
	char *filename = currprefs.qnx_keytable;
	FILE *keyfile;
	int err;
	int rawkeycode;

	prepare_keytable();
	keyfile = fopen(filename,"r");
	if (!keyfile)
		keyfile = fopen("/usr/amiga/.axlkeytable","r");
	if (!keyfile)
	{
		// this should only be a warning and using defaults.
		fprintf(stderr,"Unable to open keytable %s.\n",filename);
		return 0;
	}
	err = 0;
	while (!feof(keyfile))
	{
		if (fscanf(keyfile,"%i = ",&rawkeycode) != 1)
		{
			err = 1;
			break;
		}
		if (!scan_key(keyfile,rawkeycode,0))
		{
			err = 1;
			break;
		}
		fscanf(keyfile, " , ");
		if (!scan_key(keyfile,rawkeycode,1))
		{
			err = 1;
			break;
		}
		fscanf(keyfile, " ");
	}
	fclose(keyfile);
	if (err)
	{
		fprintf(stderr,"error in file %s\n",filename);
		return 0;
	}
	return 1;
}

static void free_keytable(void)
{
	prepare_keytable();
}

static int find_key(int *buf, int buflen)
{
	if (buflen == 1)
	{
		int rawkeycode = rawkeytable[*buf].rawkeycode;
		if (rawkeycode >= 0 && rawkeycode <= 255)
			return rawkeycode;
		else if (rawkeycode < 0)
			return -2; // sequence is not finished
	}
	else if (buflen > 1)
	{
		int index = -rawkeytable[*buf].rawkeycode;
		if (index >= 1 && index < 512)
		{
			buflen--;
			do
			{
				int nextindex = rawkeyextents[index++];
				int rawkeycode = rawkeyextents[index++];
				int numscans = rawkeyextents[index++];
				if (numscans >= buflen)
				{
					int i;

					for (i = 0; i < buflen; i++)
					{
						if (rawkeyextents[index++] != buf[i+1])
							break;
					}
					if (i == buflen)
					{
						if (i == numscans)
							return rawkeycode;
						return -2; // sequence is not finished
					}
				}
				index = nextindex;
			} while (index != 0);
		}
	}
	return -1; // not found
}

static int handle_scancode(int scancode)
{
	int r;

  //printf("scancode = 0x%02x, ",scancode);
	keybuffer[keybufferlen++] = scancode;
	while(keybufferlen > 0)
	{
		int i;
		r = find_key(keybuffer,keybufferlen);
		switch (r)
		{
			case -1:
  //printf("break sequence\n");
  				keybufferlen = 0;
  				return -1;
  				/*for(i=1; i<keybufferlen; i++)
  					keybuffer[i-1] = keybuffer[i];
				keybufferlen --;
				break;*/

			case -2:
				return -1;

			default:
  //printf("sequence finished, r = 0x%02x\n",r);
				keybufferlen = 0;
				return (r < 0 ? -1 : r);
		}
	}
	return -1;
}

/****************************************************************/

static void my_handle_events (int avoid_loop);

/* first init */
int graphics_setup (void)
{
	atexit(graphics_leave);

#ifdef SCREENBAR
	if (PtInit(NULL))
	{
		fprintf(stderr,"Can't initialize Photon\n");
		return 0;
	}
#else
	if (!PhAttach(NULL,NULL))
	{
		fprintf(stderr,"Can't attach Photon channel.\n");
		return 0;
	}
#endif

	if (!load_keytable())
		return 0;

	if (!open_region())
	{
		fprintf(stderr,"Unable to create input region\n");
		return 0;
	}

#ifdef SCREENBAR
	if (!createQnxWorkspaceBar())
	{
		fprintf(stderr,"Unable to create screenbar\n");
		return 0;
	}
	my_handle_events(1);
#endif

	return 1;
}

/* store currently set video mode from Photon - must be called before
   AGA or Picasso video mode is set */
static void store_photon_video_mode(void)
{
	if (!OldVideoModeValid)
	{
		/* PgGetVideoMode() may fail if something is still drawing */
		PgFlush();
		PgWaitDrawComplete();
		PgWaitHWIdle();
		if (PgGetVideoMode(&OldVideoMode))
		{
			fprintf(stderr,"Can't get the current video mode\n");
			exit(1);
		}
		OldVideoModeValid = 1;
	}
}

/* restore Photon video mode */
static void restore_photon_video_mode(void)
{
	if (OldVideoModeValid)
	{
		PgWaitDrawComplete();
		PgWaitHWIdle();
		PgSetVideoMode(&OldVideoMode);
		PgFlush();
		PgWaitHWIdle();
		OldVideoModeValid = 0;
	}
}

/* start direct video mode for AGA or Picasso */
static void start_direct_mode(void)
{
	PhRect_t rect;
	PhCursorInfo_t info;

#ifdef SCREENBAR
	closeQnxWorkspaceBar();
	my_handle_events(1);
#endif

	if (DirectMode == NULL)
	{
		DirectMode = PdCreateDirectContext();
		if (DirectMode == NULL)
		{
			fprintf(stderr,"Can't create direct context\n");
			exit(1);
		}
	}
	if (Olddc == NULL)
	{
		PhWindowQueryVisible(Ph_QUERY_GRAPHICS, 0, PhInputGroup(NULL), &rect);
		PhQueryCursor(PhInputGroup(NULL), &info);

		phmouse_x = info.pos.x - rect.ul.x;
		phmouse_y = info.pos.y - rect.ul.y;

		if (AmigaServerTask)
		{
			uaecptr udata;

			udata = get_long(AmigaServerTask + tc_UserData);
			if (udata)
			{
				put_long(udata + udata_new_mouse_x, phmouse_x);
				put_long(udata + udata_new_mouse_y, phmouse_y);
				uae_Signal(AmigaServerTask, SIG_SetMouse);
			}
		}

		// work-around for accidently drawn mouse pointer
		PhMoveCursorAbs(PhInputGroup(NULL), 16384, 16384);

		Olddc = PdDirectStart(DirectMode);
		if (Olddc == NULL)
		{
			fprintf(stderr,"Can't start direct mode\n");
			exit(1);
		}
		PgWaitHWIdle();
		amiga_in_front = 1;
	}
	opaque_region();
}

/* stop direct video mode for Photon */
static void stop_direct_mode(int noscreenbar)
{
	int setcursor = 0;
	PhRect_t rect;

	if (Olddc)
	{
//		PdDirectStop(DirectMode);
//		Olddc = NULL;
		amiga_in_front = 0;
		setcursor = 1;
	}
	if (DirectMode)
	{
		PdReleaseDirectContext(DirectMode);
		DirectMode = NULL;
	}
	if (Olddc)
	{
		PhDCSetCurrent(Olddc);
		Olddc = NULL;
	}
	PgWaitHWIdle();
	transparent_region();
	if (setcursor)
	{
		PhWindowQueryVisible(Ph_QUERY_GRAPHICS,0,PhInputGroup(NULL),&rect);
		if (AmigaServerTask)
		{
			uaecptr udata;

			udata = get_long(AmigaServerTask + tc_UserData);
			if (udata)
			{
				int sw,sh;
				phmouse_x = get_long(udata + udata_last_mouse_x);
				phmouse_y = get_long(udata + udata_last_mouse_y);
				sw = rect.lr.x - rect.ul.x + 1;
				sh = rect.lr.y - rect.ul.y + 1;
				if (phmouse_x < 0 || phmouse_x >= sw
					|| phmouse_y < 0 || phmouse_y >= sh)
				{
					phmouse_x = sw/2;
					phmouse_y = sh/2;
				}
			}
		}
		PhMoveCursorAbs(PhInputGroup(NULL),phmouse_x + rect.ul.x,phmouse_y + rect.ul.y);
	}
#ifdef SCREENBAR
	if (!noscreenbar)
	{
		openQnxWorkspaceBar();
		my_handle_events(1);
	}
#endif
}

/* enter graphics mode for AGA */
static void enter_graphics_mode_AGA(void)
{
	store_photon_video_mode();
	start_direct_mode();
	if (PgSetVideoMode(&DirectAGAMode))
	{
		fprintf(stderr,"Can't set AGA video mode.\n");
		exit(1);
	}
	PgFlush();
	PgWaitDrawComplete();
	PgWaitHWIdle();

	LastDirectPicassoMode = -1;
	PicassoVideoContext = NULL;
	PicassoVideoRam = NULL;

	AGAVideoRam = NULL;
	AGAVideoContext = PdCreateOffscreenContext(0, gfxvidinfo.width, gfxvidinfo.height, Pg_OSC_MAIN_DISPLAY | Pg_OSC_MEM_PAGE_ALIGN);
	if (AGAVideoContext)
	{
		AGAVideoRam = (char *) PdGetOffscreenContextPtr(AGAVideoContext);
//		gfxvidinfo.bufmem = AGAVideoRam;
//		gfxvidinfo.linemem = NULL;
		gfxvidinfo.bufmem = AGAGfxmem;
		gfxvidinfo.linemem = AGAGfxmem;
	}
	else
	{
		gfxvidinfo.bufmem = AGAGfxmem;
		gfxvidinfo.linemem = AGAGfxmem;
	}
}

/* leave AGA direct mode (switch to copy mode) */
static void leave_AGA_direct_mode(void)
{
	if (AGAVideoContext)
		PhDCRelease(AGAVideoContext);
	AGAVideoRam = NULL;
	AGAVideoContext = NULL;
	gfxvidinfo.bufmem = AGAGfxmem;
	gfxvidinfo.linemem = AGAGfxmem;
}

/* enter grahics mode for Picasso */
static void enter_graphics_mode_picasso(void)
{
	int i;
	int best_i;

	store_photon_video_mode();

/*
	printf("enter_graphics_mode_picasso: w = %d, h = %d, f = %d\n",
		picasso_vidinfo.width,picasso_vidinfo.height,picasso_vidinfo.rgbformat);
*/

	/* Check for our funny mode to display the QNX Workspace. */
	if (DirectPicassoModes[0].width == picasso_vidinfo.width
	&&  DirectPicassoModes[0].height == picasso_vidinfo.height)
	{
		leave_graphics_mode(0);
		return;
	}

	best_i = -1;
	for (i = 1; i < DirectPicassoModeNum; i++)
	{
		if (DirectPicassoModes[i].rgbformat == picasso_vidinfo.rgbformat)
		{
			if (DirectPicassoModes[i].width == picasso_vidinfo.width
			&&  DirectPicassoModes[i].height == picasso_vidinfo.height)
			{
				best_i = i;
				break;
			}
			if (DirectPicassoModes[i].width >= picasso_vidinfo.width
			&&  DirectPicassoModes[i].height >= picasso_vidinfo.height)
			{
				if (best_i == -1)
					best_i = i;
				else if (DirectPicassoModes[i].width*DirectPicassoModes[i].height
					<= DirectPicassoModes[best_i].width*DirectPicassoModes[best_i].height)
				{
					best_i = i;
				}
			}
		}
	}
	if (best_i == -1)
	{
		for (i = 0; i < DirectPicassoModeNum; i++)
		{
			if (DirectPicassoModes[i].rgbformat == picasso_vidinfo.rgbformat)
			{
				if (best_i == -1)
					best_i = i;
				else if (DirectPicassoModes[i].width*DirectPicassoModes[i].height
					>= DirectPicassoModes[best_i].width*DirectPicassoModes[best_i].height)
				{
					best_i = i;
				}
			}
		}
	}
	if (best_i != -1)
	{
		start_direct_mode();

		if (best_i == LastDirectPicassoMode)
			return;
		LastDirectPicassoMode = best_i;
		if (PgSetVideoMode(&DirectPicassoModes[best_i].mode))
		{
			fprintf(stderr,"Can't set picasso video mode.\n");
			exit(1);
		}
		PgFlush();
		PgWaitHWIdle();

		AGAVideoRam = NULL;
		AGAVideoContext = NULL;

		PicassoVideoRam = NULL;
		PicassoVideoContext = PdCreateOffscreenContext(
			0,
			DirectPicassoModes[best_i].scaledX,
			DirectPicassoModes[best_i].scaledY,
			Pg_OSC_MAIN_DISPLAY | Pg_OSC_MEM_PAGE_ALIGN);
		if (PicassoVideoContext)
		{
			PicassoVideoRam = (char *) PdGetOffscreenContextPtr(PicassoVideoContext);
			if(PicassoVideoRam)
				picasso_vidinfo.rowbytes = picasso_vidinfo.pixbytes * DirectPicassoModes[best_i].scaledX;
		}
		return;
	}

	fprintf(stderr,"Didn't find a picasso video mode.\n");
	fprintf(stderr,"picasso_vidinfo: w = %d, h = %d, d = %d, f = %d\n",
		picasso_vidinfo.width,picasso_vidinfo.height,picasso_vidinfo.depth,picasso_vidinfo.rgbformat);
	exit(1);
}

/* leave Picasso direct mode (switch to copy mode) */
static void leave_picasso_direct_mode(void)
{
	if (PicassoVideoContext)
		PhDCRelease(PicassoVideoContext);
	PicassoVideoRam = NULL;
	PicassoVideoContext = NULL;
	LastDirectPicassoMode = -1;
}

/* show picasso screen */
static void set_graphics_for_picasso(void)
{
	enter_graphics_mode_picasso();
	if (PicassoVideoRam != NULL)
		picasso_vidinfo.extra_mem = 1;
	else
		picasso_vidinfo.extra_mem = 0;
	DX_SetPalette(0,256);

	picasso_has_invalid_lines = 0;
	picasso_invalid_start = picasso_vidinfo.height + 1;
	picasso_invalid_stop = -1;
	bzero(picasso_invalid_lines, sizeof(picasso_invalid_lines));

	write_log("set_graphics_for_picasso()\n");
}

/* show AGA screen */
static void set_graphics_for_AGA(void)
{
	enter_graphics_mode_AGA();
}

/* enter graphics mode */
static void enter_graphics_mode(void)
{
	if (screen_is_picasso)
	{
		set_graphics_for_picasso();
		picasso_refresh();
	}
	else
	{
		set_graphics_for_AGA();
	}
}

/* leave graphics mode and go back to Photon */
void leave_graphics_mode(int noscreenbar)
{
	leave_AGA_direct_mode();
	leave_picasso_direct_mode();
	restore_photon_video_mode();
	stop_direct_mode(noscreenbar);
}

/* begin Amiga emulation with AGA mode */
int graphics_init (void)
{
	PgVideoModes_t ModeList;
	PgVideoModeInfo_t ModeInfo;

	int i = 0, done = 0;

	store_photon_video_mode();
	PgWaitDrawComplete(); // just to be sure
	if (PgGetVideoModeList(&ModeList))
	{
		write_log("Unable to get list of video modes.\n");
		return 0;
	}
	DirectAGAMode.refresh = 0;
	AGAVideoPixBits = 0;
	AGAVideoPixBytes = 0;
	while (i < ModeList.num_modes)
	{
		if (PgGetVideoModeInfo(ModeList.modes[i], &ModeInfo))
		{
			write_log("Can't get info for video mode %d (%d in list)\n",ModeList.modes[i],i);
			return 0;
		}
		if (ModeInfo.width == AGAWIDTH
			&& ModeInfo.height == AGAHEIGHT
			&& ModeInfo.bits_per_pixel >= 15
			&& ModeInfo.bits_per_pixel > AGAVideoPixBits)
		{
			int j = 0;

			done = 1;
			DirectAGAMode.mode = ModeList.modes[i];
			while (ModeInfo.refresh_rates[j])
			{
				if (ModeInfo.refresh_rates[j] <= OldVideoMode.refresh
					&& ModeInfo.refresh_rates[j] > DirectAGAMode.refresh)
				{
					DirectAGAMode.refresh = ModeInfo.refresh_rates[j];
				}
				j++;
			}
			AGAVideoPixBits = ModeInfo.bits_per_pixel;
			AGAVideoPixBytes = (ModeInfo.bits_per_pixel+7) / 8;
			AGAVideoRowBytes = AGAWIDTH * AGAVideoPixBytes;
		}
		i++;
	}
	if (!done)
	{
		write_log("No useful video mode found.\n");
		return 0;
	}

	gfxvidinfo.width = AGAWIDTH;
	gfxvidinfo.height = AGAHEIGHT;
	if (AGAVideoPixBytes == 3)
		gfxvidinfo.pixbytes = 4;
	else
		gfxvidinfo.pixbytes = AGAVideoPixBytes;
	gfxvidinfo.rowbytes = gfxvidinfo.width * gfxvidinfo.pixbytes;
	gfxvidinfo.maxblocklines = 0;
	AGAGfxmem = (char *) malloc(gfxvidinfo.rowbytes);
	if (!AGAGfxmem)
	{
		write_log("no memory for video buffer.\n");
		return 0;
	}
	gfxvidinfo.bufmem = AGAGfxmem;
	bzero(gfxvidinfo.bufmem,gfxvidinfo.rowbytes);
	gfxvidinfo.linemem = gfxvidinfo.bufmem;
	gfxvidinfo.emergmem = NULL;
	gfxvidinfo.can_double = 0;

	switch (AGAVideoPixBits)
	{
		case 15:
			alloc_colors64k(5,5,5,10,5,0);
			break;
		case 16:
			alloc_colors64k(5,6,5,11,5,0);
			break;
		case 24:
		case 32:
			alloc_colors64k(8,8,8,16,8,0);
			break;
	}

	enter_graphics_mode_AGA();

	buttonstate[0] = buttonstate[1] = buttonstate[2] = 0;
	lastmx = lastmy = 0;
	newmousecounters = 0;

	return 1;
}

/* stop Amiga emulation and return finally to Photon */
void graphics_leave (void)
{
	leave_graphics_mode(1);
	close_region();
#ifdef SCREENBAR
	destroyQnxWorkspaceBar();
#endif
	free_keytable();
}

/* called in custom reset */
void graphics_reset (void)
{
	AmigaServerTask = NULL;
}

/* handle all that input stuff and other events that has to
   happen each "video frame" */
static void
my_handle_events (int avoid_loop)
{
	PhRawKeyEvent_t *key;
	PhRawPtrEvent_t *ptr;
	PhSystemEvent_t *sys;
	int i, r, rcode, autorepeat;
	int record_this_key;
	newmousecounters = 0;

	while ((r = PhEventPeek(event,eventsize)))
	{
		switch (r)
		{
			case Ph_EVENT_MSG:
				if (event->collector.rid == input_region)
				{
					switch (event->type)
					{
						case Ph_EV_RAW:
							switch (event->subtype)
							{
								case Ph_EV_RAW_PTR:
									ptr = (PhRawPtrEvent_t *) PhGetData(event);
									for(i=0; i<ptr->num_coord; i++)
									{
										// for wheel-mouse support
										if(ptr->coord[i].z > 0)
										{
											if(keystate[0x4d])
											{
												keystate[0x4d] = 0;
												record_key((0x4d << 1) | 1);
											}
											qnxkeystate[0x4d] = 0;
										}
										else if(ptr->coord[i].z < 0)
										{
											if(keystate[0x4c])
											{
												keystate[0x4c] = 0;
												record_key((0x4c << 1) | 1);
											}
											qnxkeystate[0x4c] = 0;
										}
									}
									if (amiga_in_front)
									{
										if (ptr->num_coord > 0)
										{
											if (ptr->raw_flags & Ph_PTR_RAW_FLAG_XY_REL)
											{
												for (i = 0; i < ptr->num_coord; i++)
												{
													lastmx += ptr->coord[i].x;
													lastmy -= ptr->coord[i].y;
												}
											}
											else
											{
												lastmx = ptr->coord[ptr->num_coord-1].x;
												lastmy = ptr->coord[ptr->num_coord-1].y;
											}
										}
										if (!(ptr->raw_flags & Ph_PTR_RAW_FLAG_NO_BUTTONS))
										{
											buttonstate[0] = (ptr->button_state & Ph_BUTTON_SELECT) != 0;
											buttonstate[1] = (ptr->button_state & Ph_BUTTON_ADJUST) != 0;
											buttonstate[2] = (ptr->button_state & Ph_BUTTON_MENU) != 0;
										}
									}
									else
									{
										// Avoid Blanker to pop up.
										static int counter = 10;
										if(counter < 0)
										{
											record_key((0x60 << 1) | 1);
											counter = 10;
										}
										else
										{
											counter --;
										}
									}
									break;
								case Ph_EV_RAW_KEY:
									key = (PhRawKeyEvent_t *) PhGetData(event);
									record_this_key = 0;
									rcode = 0;
									autorepeat = (key->key_flags & Pk_KF_Key_Repeat) != 0;
									if (key->key_flags & Pk_KF_Scan_Valid)
									{
										static int shift_lock_down = 0;
										int key_handled = 0;
										if(key->key_mods & Pk_KM_Shift_Lock)
										{
											if(!shift_lock_down)
											{
												rcode = 0x62;
												keystate[rcode] = 1;
												record_key(rcode << 1);
												shift_lock_down = 1;
												key_handled = 1;
											}
										}
										else
										{
											if(shift_lock_down)
											{
												rcode = 0x62;
												keystate[rcode] = 0;
												record_key((rcode << 1) | 1);
												shift_lock_down = 0;
												key_handled = 1;
											}
										}
										rcode = handle_scancode(key->key_scan);
										if (rcode != -1  &&  (rcode & 0x7f) != 0x62)
										{
											if (rcode & 0x80)
											{
												/* KEY UP */
												rcode &= 0x7F;
												if(keystate[rcode])
												{
													keystate[rcode] = 0;
													record_key((rcode << 1) | 1);
													key_handled = 1;
												}
												qnxkeystate[rcode] = 0;
											}
											else
											{
												/* record keys if amiga screen in front
												   or left amiga key is pressed */
												if (amiga_in_front || rcode == 0x66)
												{
													if (!keystate[rcode])
													{
														keystate[rcode] = 1;
														record_this_key = 1;
														key_handled = 1;
													}
												}
												if (!qnxkeystate[rcode])
													qnxkeystate[rcode] = 1;
											}
										}
										if(!amiga_in_front  &&  !key_handled)
										{
											// Avoid Blanker to pop up.
											record_key((0x60 << 1) | 1);
										}
									}

									if (!autorepeat)
									{
										// evaluate meta key strokes
										if ((keystate[0x63] && keystate[0x66] && keystate[0x67])
											|| (qnxkeystate[0x63] && qnxkeystate[0x66] && qnxkeystate[0x67]))
										{
											// Ctrl-Amiga-Amiga
											uae_reset();
											//system("shutdown");
											//if (amiga_in_front)
											//	leave_graphics_mode(1);
											//activate_debugger();
										}
										else if ((keystate[0x66] && keystate[0x67] && keystate[0x45])
											|| (qnxkeystate[0x66] && qnxkeystate[0x67] && qnxkeystate[0x45]))
										{
											// LeftAmiga-RightAmiga-Escape
											uae_quit();
										}
										else if (keystate[0x66] && keystate[0x32])
										{
											// LeftAmiga-X
											if (amiga_in_front)
											{
												if (AmigaServerTask)
													uae_Signal(AmigaServerTask,SIG_AmigaToBack);
												else
												{
													if (!avoid_loop)
														leave_graphics_mode(0);
												}
											}
											record_this_key = 0;
										}
										else if (qnxkeystate[0x66] && qnxkeystate[0x32])
										{
											// LeftAmiga-X
											if (!amiga_in_front)
											{
												if (AmigaServerTask)
													uae_Signal(AmigaServerTask,SIG_AmigaToFront);
												else
												{
													if (!avoid_loop)
														enter_graphics_mode();
												}
											}
											record_this_key = 0;
										}
										else if (qnxkeystate[0x66] && (qnxkeystate[0x36] || qnxkeystate[0x37]))
										{
											// LeftAmiga-N or LeftAmiga-M
											record_key(0x66 << 1);
											if (qnxkeystate[0x36])
												record_key(0x36 << 1);
											if (qnxkeystate[0x37])
												record_key(0x37 << 1);
											if (qnxkeystate[0x36])
												record_key(0x36 << 1 | 1);
											if (qnxkeystate[0x37])
												record_key(0x37 << 1 | 1);
											record_key(0x66 << 1 | 1);
											record_this_key = 0;
											if (!AmigaServerTask && !avoid_loop)
												enter_graphics_mode();
										}
									}
									if (record_this_key)
									{
										record_key(rcode << 1);
									}
									break;
								default:
									break;
							}
							break;
						case Ph_EV_SYSTEM:
							if (event->subtype == Ph_SYSTEM_REGION_CHANGE)
							{
								sys = (PhSystemEvent_t *) PhGetData(event);
								if (sys->RegionChange.fields == 0xFFFFFFFF)
								{
									if (amiga_in_front)
									{
										if (AmigaServerTask)
										{
											uae_Signal(AmigaServerTask,SIG_AmigaToBack);
										}
										else
										{
											if (!avoid_loop)
												leave_graphics_mode(0);
										}
									}
								}
							}
							break;
						default:
							break;
					}
				}
				else
				{
#ifdef SCREENBAR
					PtEventHandler(event);
					if (qnxWorkspaceToBackPressed)
					{
						qnxWorkspaceToBackPressed = 0;
						if (!amiga_in_front)
						{
							// enter_graphics_mode();
							record_key(0x66 << 1);
							record_key(0x37 << 1);
							record_key(0x37 << 1 | 1);
							record_key(0x66 << 1 | 1);
							if (!AmigaServerTask && !avoid_loop)
								enter_graphics_mode();
						}
					}
#endif
				}
				break;
			case Ph_RESIZE_MSG:
				write_log("resize event buffer\n");
				break;
			case -1:
				write_log("error from event peek\n");
				break;
		}
	}

	if (screen_is_picasso
	&&  picasso_display_state
	&&  !picasso_vidinfo.extra_mem
	&&  picasso_has_invalid_lines)
	{
		int i;
		int strt = -1;

		if(picasso_invalid_stop >= picasso_vidinfo.height)
			picasso_invalid_stop = picasso_vidinfo.height - 1;

		picasso_invalid_lines[picasso_vidinfo.height] = 0;
		for (i = picasso_invalid_start; i < picasso_invalid_stop + 2; i++)
		{
			if (picasso_invalid_lines[i])
			{
				picasso_invalid_lines[i] = 0;
				if (strt != -1)
					continue;
				strt = i;
			}
			else
			{
				if (strt == -1)
					continue;
				flush_block_picasso(strt, i-1);
				strt = -1;
			}
		}
		if (strt != -1)
		{
			fprintf(stderr,"problem with refreshing invalid lines\n");
			uae_quit();
		}
		picasso_has_invalid_lines = 0;
		picasso_invalid_start = picasso_vidinfo.height + 1;
		picasso_invalid_stop = -1;
	}
}

void handle_events(void)
{
	my_handle_events(0);
}

int check_prefs_changed_gfx (void)
{
	if(changed_prefs.gfx_linedbl != currprefs.gfx_linedbl)
	{
		currprefs.gfx_linedbl = changed_prefs.gfx_linedbl;
		return 1;
	}
	return 0;
}

int debuggable (void)
{
	return 0;
}

int needmousehack (void)
{
	return 0;
}

void LED (int on)
{
}

void DX_Invalidate (int first, int last)
{
	if(picasso_vidinfo.extra_mem != 0)
		return;
	if (first > last)
		return;

	//printf("DX_Invalidate: first = %d, last = %d\n", first, last);

	picasso_has_invalid_lines = 1;
	if (first < picasso_invalid_start)
		picasso_invalid_start = first;
	if (last > picasso_invalid_stop)
		picasso_invalid_stop = last;

	while (first <= last)
	{
		picasso_invalid_lines[first] = 1;
		first++;
	}
}

int DX_BitsPerCannon (void)
{
	return 8;
}

static PgColor_t clut[256];
static int clutchanged = 0;

void DX_SetPalette (int start, int count)
{
	if (screen_is_picasso && picasso_vidinfo.pixbytes == 1 && count > 0)
	{
		struct MyCLUTEntry *c = &picasso96_state.CLUT[start];
		clutchanged = 1;
		do
		{
			clut[start++] = PgRGB(c->Red,c->Green,c->Blue);
			c++;
		} while (--count > 0);
	}
}

void DX_SetPalette_vsync (void)
{
	if (clutchanged)
	{
		PgSetPalette(clut,0,0,256,Pg_PALSET_HARDLOCKED,0);
		PgFlush();
		clutchanged = 0;
	}
}

static uae_u32 get_format(unsigned long type)
{
	uae_u32 r = (uae_u32) -1;
	switch (type)
	{
		case Pg_IMAGE_PALETTE_BYTE:
			r = RGBFB_CLUT;
			break;
		case Pg_IMAGE_DIRECT_8888:
			r = RGBFB_B8G8R8A8;
			break;
		case Pg_IMAGE_DIRECT_888:
			r = RGBFB_B8G8R8;
			break;
		case Pg_IMAGE_DIRECT_555:
			r = RGBFB_R5G5B5PC;
			break;
		case Pg_IMAGE_DIRECT_565:
			r = RGBFB_R5G6B5PC;
			break;
	}
	return r;
}

static uae_u8 getbytesperpixel(uae_u32 RGBfmt)
{
	switch (RGBfmt)
	{
		case RGBFB_CLUT:
			return 1;

		case RGBFB_A8R8G8B8:
		case RGBFB_A8B8G8R8:
		case RGBFB_R8G8B8A8:
		case RGBFB_B8G8R8A8:
			return 4;

		case RGBFB_B8G8R8:
		case RGBFB_R8G8B8:
			return 3;

		case RGBFB_R5G5B5:
		case RGBFB_R5G6B5:
		case RGBFB_R5G6B5PC:
		case RGBFB_R5G5B5PC:
		case RGBFB_B5G6R5PC:
		case RGBFB_B5G5R5PC:
			return 2;

		default:
			write_log ("ERROR - GetBytesPerPixel() was unsuccessful with 0x%x?!\n", RGBfmt);
	}
	return 0;
}

int DX_FillResolutions (uae_u16 *ppixel_format)
{
	#define NUMSPCRES 4
	PgVideoModes_t ModeList;
	PgVideoModeInfo_t ModeInfo;
	int i = 0, count = 0, j, k;
	const int sHeight[NUMSPCRES] = { 200, 240, 256, 400 };
	const int sWidth[NUMSPCRES]  = { 320, 320, 320, 640 };
	int sOk[NUMSPCRES] = { 0, 0, 0, 0 };
	int refreshrate;
	uae_u32 format;

	PgWaitDrawComplete(); // just to be sure...
	if (PgGetVideoModeList(&ModeList))
	{
		write_log("Unable to get list of video modes.\n");
		return 0;
	}

	*ppixel_format = 0;
	DirectPicassoModeNum = 0;

	// funny mode for automatic Photon/Amiga switch
	DisplayModes[count].res.width = 320;
	DisplayModes[count].res.height = 16;
	DisplayModes[count].depth = 1;
	DisplayModes[count].refresh = 60;
	DisplayModes[count].name[0] = 0;
	*ppixel_format |= 1 << RGBFB_CLUT;
	count++;
	DirectPicassoModes[DirectPicassoModeNum].mode.mode = -1;
	DirectPicassoModes[DirectPicassoModeNum].mode.refresh = 0;
	DirectPicassoModes[DirectPicassoModeNum].width = 320;
	DirectPicassoModes[DirectPicassoModeNum].height = 16;
	DirectPicassoModes[DirectPicassoModeNum].depth = 8;
	DirectPicassoModes[DirectPicassoModeNum].rgbformat = RGBFB_CLUT;
	DirectPicassoModes[DirectPicassoModeNum].scaledX = 320;
	DirectPicassoModes[DirectPicassoModeNum].scaledY = 16;
	DirectPicassoModeNum++;

	i = 0;
	while (i < ModeList.num_modes)
	{
		if (PgGetVideoModeInfo(ModeList.modes[i], &ModeInfo))
		{
			write_log("Can't get info for video mode %d (%d in list)\n",ModeList.modes[i],i);
			return 0;
		}
		if (ModeInfo.bytes_per_scanline*ModeInfo.height <= (int) allocated_gfxmem
		&& (format = get_format(ModeInfo.type)) != (uae_u32) -1)
		{
			refreshrate = 0;
			if (OldVideoModeValid)
			{
				for(j=0; ModeInfo.refresh_rates[j]; j++)
				{
					if (ModeInfo.refresh_rates[j] <= OldVideoMode.refresh
					&&  ModeInfo.refresh_rates[j] > refreshrate)
					{
						refreshrate = ModeInfo.refresh_rates[j];
					}
				}
			}
			for (j=0; j<NUMSPCRES; j++)
			{
				if(sWidth[j]  == ModeInfo.width
				&& sHeight[j] == ModeInfo.height)
				{
					sOk[j] = 1;
					break;
				}
			}
			write_log("add mode 0x%08x, %dx%dx%d, %d Hz, format = %d\n",ModeList.modes[i], ModeInfo.width,ModeInfo.height,ModeInfo.bits_per_pixel,refreshrate,format);
			DisplayModes[count].res.width = ModeInfo.width;
			DisplayModes[count].res.height = ModeInfo.height;
			DisplayModes[count].depth = getbytesperpixel(format);
			DisplayModes[count].refresh = refreshrate ? refreshrate : 60;
			DisplayModes[count].name[0] = 0;
			*ppixel_format |= 1 << format;
			count++;
			DirectPicassoModes[DirectPicassoModeNum].mode.mode = ModeList.modes[i];
			DirectPicassoModes[DirectPicassoModeNum].mode.refresh = refreshrate;
			DirectPicassoModes[DirectPicassoModeNum].width = ModeInfo.width;
			DirectPicassoModes[DirectPicassoModeNum].height = ModeInfo.height;
			DirectPicassoModes[DirectPicassoModeNum].depth = getbytesperpixel(format)*8;
			DirectPicassoModes[DirectPicassoModeNum].rgbformat = format;
			DirectPicassoModes[DirectPicassoModeNum].scaledX = ModeInfo.width;
			DirectPicassoModes[DirectPicassoModeNum].scaledY = ModeInfo.height;
			DirectPicassoModeNum++;
			if (DirectPicassoModeNum >= PICASSO_MAX_MODES)
				break;
		}
		i++;
	}
	for(j=0; j<NUMSPCRES; j++)
	{
		if(!sOk[j])
		{
			int bestWidth = 999999;
			int bestHeight = 999999;

			for(i=0; i<ModeList.num_modes; i++)
			{
				if(PgGetVideoModeInfo(ModeList.modes[i], &ModeInfo) == 0)
				{
					if(ModeInfo.width  >= sWidth[j]
					&& ModeInfo.height >= sHeight[j])
					{
						if(ModeInfo.width  % sWidth[j]  == 0
						&& ModeInfo.height % sHeight[j] == 0
						&& ModeInfo.width  / sWidth[j]  <= 2
						&& ModeInfo.height / sHeight[j] <= 2)
						{
							if(bestWidth % sWidth[j] != 0
							|| bestHeight % sHeight[j] != 0
							|| ModeInfo.width < bestWidth
							|| ModeInfo.height < bestHeight)
							{
								bestWidth = ModeInfo.width;
								bestHeight = ModeInfo.height;
							}
						}
						else if((bestWidth % sWidth[j] != 0  ||  bestHeight % sHeight[j] != 0)
						&& (ModeInfo.width < bestWidth || ModeInfo.height < bestHeight))
						{
							bestWidth = ModeInfo.width;
							bestHeight = ModeInfo.height;
						}
						else if(bestWidth == 999999)
						{
							bestWidth = ModeInfo.width;
							bestHeight = ModeInfo.height;
						}
					}
				}
			}
			if(bestWidth < 999999)
			{
				for(i=0; i<ModeList.num_modes; i++)
				{
					if(PgGetVideoModeInfo(ModeList.modes[i], &ModeInfo) == 0)
					{
						if(ModeInfo.width == bestWidth
						&& ModeInfo.height == bestHeight
						&& ModeInfo.bytes_per_scanline * ModeInfo.height <= (int) allocated_gfxmem
						&& (format = get_format(ModeInfo.type)) != (uae_u32) -1)
						{
							refreshrate = 0;
							if (OldVideoModeValid)
							{
								for(k=0; ModeInfo.refresh_rates[k]; k++)
								{
									if (ModeInfo.refresh_rates[k] <= OldVideoMode.refresh
									&&  ModeInfo.refresh_rates[k] > refreshrate)
									{
										refreshrate = ModeInfo.refresh_rates[k];
									}
								}
							}
							if (DirectPicassoModeNum >= PICASSO_MAX_MODES)
								break;
							write_log("add scaled mode 0x%08x, %dx%dx%d, %d Hz, format = %d\n",ModeList.modes[i], sWidth[j], sHeight[j], ModeInfo.bits_per_pixel, refreshrate, format);
							DisplayModes[count].res.width  = sWidth[j];
							DisplayModes[count].res.height = sHeight[j];
							DisplayModes[count].depth = getbytesperpixel(format);
							DisplayModes[count].refresh = refreshrate ? refreshrate : 60;
							DisplayModes[count].name[0] = 0;
							count++;
							DirectPicassoModes[DirectPicassoModeNum].mode.mode = ModeList.modes[i];
							DirectPicassoModes[DirectPicassoModeNum].mode.refresh = refreshrate;
							DirectPicassoModes[DirectPicassoModeNum].width  = sWidth[j];
							DirectPicassoModes[DirectPicassoModeNum].height = sHeight[j];
							DirectPicassoModes[DirectPicassoModeNum].depth = getbytesperpixel(format)*8;
							DirectPicassoModes[DirectPicassoModeNum].rgbformat = format;
							DirectPicassoModes[DirectPicassoModeNum].scaledX = ModeInfo.width;
							DirectPicassoModes[DirectPicassoModeNum].scaledY = ModeInfo.height;
							DirectPicassoModeNum++;
						}
					}
				}
			}
		}
	}
	return count;
}

void gfx_set_picasso_modeinfo (int w, int h, int depth, int rgbfmt)
{
	picasso_vidinfo.width = w;
	picasso_vidinfo.height = h;
	picasso_vidinfo.depth = depth;
	picasso_vidinfo.pixbytes = getbytesperpixel(rgbfmt);
	if(LastDirectPicassoMode >= 0  &&  DirectPicassoModes[LastDirectPicassoMode].width == w)
		picasso_vidinfo.rowbytes = picasso_vidinfo.pixbytes * DirectPicassoModes[LastDirectPicassoMode].scaledX;
	else
		picasso_vidinfo.rowbytes = picasso_vidinfo.pixbytes * w;
	picasso_vidinfo.rgbformat = rgbfmt;
	if (screen_is_picasso)
		set_graphics_for_picasso();
}

void gfx_set_picasso_baseaddr (uaecptr a)
{
}

void gfx_set_picasso_state (int on)
{
	if (on == screen_is_picasso)
		return;
	screen_is_picasso = on;
	if (on)
	{
		set_graphics_for_picasso();
	}
	else
	{
		set_graphics_for_AGA();
	}
}

uae_u8 *gfx_lock_picasso (void)
{
	return PicassoVideoRam;
}

void gfx_unlock_picasso (void)
{
}

int lockscr (void)
{
	return 1;
}

void unlockscr (void)
{
}

void target_save_options(FILE *f, struct uae_prefs *p)
{
	fprintf(f,"qnx.keytable=%s\n",p->qnx_keytable);
}

int target_parse_option (struct uae_prefs *p, char *option, char *value)
{
	return (cfgfile_string(option,value,"keytable",p->qnx_keytable,sizeof(p->qnx_keytable)));
}


int NDX_InvertRect(struct RenderInfo* ri, unsigned long X, unsigned long Y, unsigned long Width, unsigned long Height, uae_u32 mask, int Bpp)
{
    return 0;
}

int NDX_BlitPattern(struct RenderInfo* ri,struct Pattern* pat,unsigned long X, unsigned long Y, unsigned long W, unsigned long H, uae_u8 Mask, uae_u32 RGBFmt)
{
    return 0;
}

int NDX_BlitTemplate(struct RenderInfo* ri, struct Template* tmp, unsigned long X, unsigned long Y, unsigned long W, unsigned long H, uae_u16 Mask)
{
    return 0;
}

int NDX_BlitPlanar2Chunky(struct RenderInfo* ri, struct BitMap* bm, unsigned long srcx, unsigned long srcy, unsigned long dstx, unsigned long dsty, unsigned long width, unsigned long height, uae_u8 minterm, uae_u8 mask)
{
    return 0;
}

int NDX_BlitPlanar2Direct(struct RenderInfo* ri, struct BitMap* bm, unsigned long srcx, unsigned long srcy, unsigned long dstx, unsigned long dsty, unsigned long width, unsigned long height, uae_u8 minterm, uae_u8 Mask)
{
    return 0;
}

int NDX_FillRect(struct RenderInfo* ri, unsigned long X, unsigned long Y, unsigned long Width, unsigned long Height, uae_u32 Pen, uae_u8 Mask, uae_u32 RGBFormat)
{
    return 0;
}

int NDX_BlitRect(struct RenderInfo* ri, unsigned long srcx, unsigned long srcy, unsigned long dstx, unsigned long dsty, unsigned long width, unsigned long height, uae_u8 Mask)
{
    return 0;
}

int NDX_BlitRectNoMaskComplete(struct RenderInfo* sri,struct RenderInfo* dri, unsigned long srcx, unsigned long srcy, unsigned long dstx, unsigned long dsty, unsigned long width, unsigned long height, uae_u8 OpCode, uae_u32 RGBFmt)
{
    return 0;
}
