 /*
  * UAE - The Un*x Amiga Emulator
  *
  * Support for the ALSA audio interface
  *
  * Copyright 2001 HAAGE&PARTNER Computer GmbH
  */

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"
#include "memory.h"
#include "custom.h"
#include "audio.h"
#include "gensound.h"
#include "sounddep/sound.h"
#include "threaddep/penguin.h"
#include <sys/asoundlib.h>
#include <math.h>

#ifndef __FUNC__
#define __FUNC__ __PRETTY_FUNCTION__
#endif


snd_pcm_t *audio_handle = NULL;
static int audio_card;
static int audio_device;

int audio_buffer_size;
int audio_fragmt_size;
char *audio_buffers[4] = { NULL, NULL, NULL, NULL };
char *audio_buffer_walk;

smp_comm_pipe audio_pipe;
penguin_id audio_thread_id = 0;


static void playA1000Melody(double basefreq, int channels, int bytesPerSample)
{
#define NUMTONES 10
#define DURATION 0.9
#define M_2PI (2.0 * M_PI)
	static double freq[] = { 277.2, 415.3, 349.3, 277.2, 311.2, 349.3, 370.0, 349.3, 311.2, 311.2 };
	int numSamples;
	char *waveform;

	numSamples = channels * ((int)(basefreq * DURATION));
	waveform = (char*) malloc(numSamples * bytesPerSample);

	if(waveform)
	{
		int s, l, t, c;

		if(bytesPerSample == 1)
		{
			char *walk = waveform;
			char v = 0;
			double w = 0.0, aw = 0.0;

			for(t=0; t<NUMTONES; t++)
			{
				l = (int)(DURATION * basefreq / NUMTONES) - 10;
				aw = M_2PI * freq[t] / basefreq;
				for(s=0; s<l; s++)
				{
					if(walk >= waveform + numSamples * bytesPerSample)
						break;
					v =  (char)(cos(w) * 12.0);
					for(c = channels; c>0; c--)
						*walk++ = v;
					w += aw;
					while(w > M_2PI)
						w -= M_2PI;
				}
				for(s=0; s<10; s++)
				{
					if(walk >= waveform + numSamples * bytesPerSample)
						break;
					for(c = channels; c>0; c--)
						*walk++ = v;
				}
			}
		}
		else if(bytesPerSample == 2)
		{
			short *walk = (short*) waveform;
			short v = 0;
			double w = 0.0, aw = 0.0;

			for(t=0; t<NUMTONES; t++)
			{
				l = (int)(DURATION * basefreq / NUMTONES) - 10;
				aw = M_2PI * freq[t] / basefreq;
				for(s=0; s<l; s++)
				{
					if(walk >= (short*)(waveform + numSamples * bytesPerSample))
						break;
					v =  (short)(cos(w) * 3000.0);
					for(c = channels; c>0; c--)
						*walk++ = v;
					w += aw;
					while(w > M_2PI)
						w -= M_2PI;
				}
				for(s=0; s<10; s++)
				{
					if(walk >= (short*)(waveform + numSamples * bytesPerSample))
						break;
					for(c = channels; c>0; c--)
						*walk++ = v;
				}
			}
		}

		snd_pcm_plugin_write(audio_handle, waveform, numSamples * bytesPerSample);
		snd_pcm_plugin_flush(audio_handle, SND_PCM_CHANNEL_PLAYBACK);
		free(waveform);
	}
}


static void *audio_thread(void *arg)
{
	char *b;
	int r, l;
	static int melodyOnce = 0;

	if(!melodyOnce)
	{
		playA1000Melody(currprefs.sound_freq, currprefs.stereo ? 2 : 1, currprefs.sound_bits / 8);
		while(comm_pipe_has_data(&audio_pipe))
			read_comm_pipe_pvoid_blocking(&audio_pipe);
		melodyOnce = 1;
	}

	while(1)
	{
		b = (char*) read_comm_pipe_pvoid_blocking(&audio_pipe);
		if(!b)
			pthread_exit(NULL);

		while(comm_pipe_has_data(&audio_pipe))
		{
			l ++;
			b = (char*) read_comm_pipe_pvoid_blocking(&audio_pipe);
			if(!b)
				pthread_exit(NULL);
		}

		l = audio_buffer_size;
		while(l > 0)
		{
			r = snd_pcm_plugin_write(audio_handle, b, l);
			if(r > 0)
			{
				b += r;
				l -= r;
			}
			else if(r == -EAGAIN)
			{
				write_log("audio_thread: error EAGAIN.\n");
				break;
			}
			else
			{
				snd_pcm_channel_status_t status;

				write_log("audio_thread: error on writing.\n");
				bzero(&status, sizeof(status));

				status.channel = SND_PCM_CHANNEL_PLAYBACK;
				if(snd_pcm_plugin_status(audio_handle, &status) != 0)
				{
					write_log("audio_thread: snd_pcm_channel_status() failed.\n");
					break;
				}
				if(status.status == SND_PCM_STATUS_READY  ||  status.status == SND_PCM_STATUS_UNDERRUN)
				{
					write_log("audio_thread: repreparing...\n");
					if(snd_pcm_plugin_prepare(audio_handle, SND_PCM_CHANNEL_PLAYBACK) != 0)
					{
						write_log("failed.\n");
						break;
					}
				}
				else
				{
					write_log("audio_thread: illegal state = %d.\n", status.status);
					break;
				}
			}
		}
	}
	return NULL;
}


int setup_sound (void)
{
	sound_available = 1;
    sample_evtime = (long)maxhpos * maxvpos * 50 / currprefs.sound_freq;
    return 1;
}


int init_sound (void)
{
	int i;
	snd_pcm_channel_params_t params;
	snd_pcm_channel_info_t info;

	write_log("%s: entering ...\n", __FUNC__);

	bzero(&params, sizeof(params));
	bzero(&info, sizeof(info));

	if(snd_pcm_open_preferred(&audio_handle, &audio_card, &audio_device, SND_PCM_OPEN_PLAYBACK | SND_PCM_OPEN_NONBLOCK) != 0)
	{
		write_log("Failed to connect to audio interface.\n");
		sound_available = 0;
		return 0;
	}
	snd_pcm_nonblock_mode(audio_handle, 0);

	info.channel = SND_PCM_CHANNEL_PLAYBACK;
	if(snd_pcm_plugin_info(audio_handle, &info) != 0)
	{
		write_log("Failed to obtain channel info from the audio interface.\n");
		sound_available = 0;
		return 0;
	}
	write_log("supports block mode: %s\n", info.flags & SND_PCM_CHNINFO_BLOCK ? "Yes" : "*NO*");
	write_log("supports stream mode: %s\n", info.flags & SND_PCM_CHNINFO_STREAM ? "Yes" : "*NO*");
	write_log("min frequency: %d\n", info.min_rate);
	write_log("max frequency: %d\n", info.max_rate);
	write_log("min voices: %d\n", info.min_voices);
	write_log("max voices: %d\n", info.max_voices);
	write_log("max buffer size: %d\n", info.max_buffer_size);
	write_log("min fragment size: %d\n", info.min_fragment_size);
	write_log("max fragment size: %d\n", info.max_fragment_size);

	if(currprefs.stereo  &&  info.max_voices < 2)
	{
		write_log("Stereo not supported.\n");
		return 0;
	}
	else if(!currprefs.stereo  &&  info.min_voices > 1)
	{
		write_log("Mono not supported.\n");
		return 0;
	}

	if((currprefs.sound_bits ==  8  &&  !(info.formats & SND_PCM_FMT_S8 ))
	|| (currprefs.sound_bits == 16  &&  !(info.formats & SND_PCM_FMT_S16))
	|| (currprefs.sound_bits < 8)
	|| (currprefs.sound_bits > 16)
	|| (currprefs.sound_bits & 7))
	{
		write_log("%d bits per sample not supported.\n", currprefs.sound_bits);
		return 0;
	}

	if(currprefs.sound_freq < info.min_rate)
	{
		write_log("Samplerate not supported. Minimum rate is %d Hz.\n", info.min_rate);
		return 0;
	}
	else if(currprefs.sound_freq > info.max_rate)
	{
		write_log("Samplerate not supported. Maximum rate is %d Hz.\n", info.max_rate);
		return 0;
	}

	audio_buffer_size = currprefs.sound_minbsiz;
	if(audio_buffer_size < info.min_fragment_size)
		audio_buffer_size = info.min_fragment_size;
	else if(audio_buffer_size > info.max_fragment_size)
		audio_buffer_size = info.max_fragment_size;
	audio_fragmt_size = audio_buffer_size;
	write_log("audio_buffer_size = %d, audio_fragmt_size = %d, freq = %d Hz\n", audio_buffer_size, audio_fragmt_size, currprefs.sound_freq);

	params.channel = SND_PCM_CHANNEL_PLAYBACK;
	params.mode = SND_PCM_MODE_BLOCK;
	params.format.interleave = currprefs.stereo ? 1 : 0;
	params.format.format = (currprefs.sound_bits == 16 ? SND_PCM_SFMT_S16 : SND_PCM_SFMT_S8);
	params.format.rate = currprefs.sound_freq;
	params.format.voices = currprefs.stereo ? 2 : 1;
	params.start_mode = SND_PCM_START_DATA;
	params.stop_mode = SND_PCM_STOP_STOP;
	params.buf.block.frag_size = audio_fragmt_size;
	params.buf.block.frags_min = 1;
	params.buf.block.frags_max = 2;
	if(snd_pcm_plugin_params(audio_handle, &params) != 0)
	{
		write_log("Failed to configure audio interface: %d.\n", params.why_failed);
		return 0;
	}

	for(i=0; i<4; i++)
	{
		audio_buffers[i] = (char*) malloc(audio_buffer_size);
		if(!audio_buffers[i])
		{
			write_log("Failed to allocate %d bytes for playback buffer.\n", audio_buffer_size);
			return 0;
		}
		bzero(audio_buffers[i], audio_buffer_size);
	}

	if(snd_pcm_plugin_prepare(audio_handle, SND_PCM_CHANNEL_PLAYBACK) != 0)
	{
		write_log("Failed to prepare audio interface.\n");
		return 0;
	}

    sample_evtime = (long)maxhpos * maxvpos * 50 / currprefs.sound_freq;
    if (currprefs.sound_bits == 16)
    {
		init_sound_table16 ();
		sample_handler = currprefs.stereo ? sample16s_handler : sample16_handler;
    }
    else
    {
		init_sound_table8 ();
		sample_handler = currprefs.stereo ? sample8s_handler : sample8_handler;
    }

	audio_buffer_walk = audio_buffers[0];
	init_comm_pipe(&audio_pipe, 64, 1);
	if(start_penguin(audio_thread, NULL, &audio_thread_id) != EOK)
	{
		write_log("Failed to launch audio thread.\n");
		return 0;
	}

	sound_available = 1;
	write_log("%s: success.\n", __FUNC__);
    return 1;
}


void close_sound(void)
{
	int i;
	write_log("%s\n", __FUNC__);

	if(audio_thread_id)
	{
		write_comm_pipe_pvoid(&audio_pipe, NULL, 1);
		while(pthread_join(audio_thread_id, NULL) == EINTR)
			;
		audio_thread_id = 0;
	}
	if(audio_handle)
	{
		snd_pcm_close(audio_handle);
		audio_handle = NULL;
	}
	for(i=0; i<4; i++)
	{
		free(audio_buffers[i]);
		audio_buffers[i] = NULL;
	}
}
